/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

# include <stdlib.h>
# include <ulocks.h>
# include <sys/types.h>
# include <sys/prctl.h>
# include <sys/schedctl.h>

int ROOT_ID=1;
static int j=0;

# include "ukey.h"
# include "barrier.h"
# include "bcast.h"
# include "anl_shmem.h"
# include "mbuffer.h"
# include "mqueue.h"
# include "twait.h"
# include "struct.h"

#ifdef MPISIM
# include "sim.h"
#endif

/* each thread can have 2M shared memory for message handles and other system objects */
/* such as barriers, broadcast objects and 6M for system buffering. */
/* This will allow about 20000 pending messages for each thread, should be large enough. */
/**
# define GLOBAL_SHARE (0x800000) 
# define SYS_BUFFER (GLOBAL_SHARE - 0x200000)
**/
# define GLOBAL_SHARE (0x1000000) 
# define SYS_BUFFER (GLOBAL_SHARE - 0x200000)

/* definitions for global data structures, shared by all threads */
int	thrMPI_Nproc=0; /* number of processors */

/* shared arena pointer */
usptr_t *thrMPI_sc_handle;

/* referenced in context/comm.c */
thrMPI_Comm **thrMPI_COMM_WORLD;
thrMPI_Bar *MPI_COMM_WORLD_BAR;
thrMPI_BcastObj *MPI_COMM_WORLD_BC;
thrMPI_ReduceObj *MPI_COMM_WORLD_RDC;
thrMPI_AllreduceObj *MPI_COMM_WORLD_ARDC;

/* static variables */
static int ext_argc;
static char **ext_argv;

/* prototype */
void thrMainInit(void); /* global init */
static void thrMPI_Startup(void); /* our startup stub */

static void uexit(int exit_code)
{
	anl_shmem_gend();
	thrMPI_exit(exit_code);
}

/* the main entrance of the MPI program */
main(int argc, char *argv[])
{
	int np; /* number of processors */
	char *MPI_NPROC=NULL; /* environment variables */

# if defined(PRC)
	int knp;
	char *SYS_NPROC=NULL;
# endif

  	ukey_thr_t*	Threads;
  	int i;

	/* save back argv and argc */
	ext_argc=argc;
	ext_argv=argv;

	/* MPI_NPROC will be set to be #proc in tmpirun */
	MPI_NPROC=getenv("MPI_NPROC"); 

# if defined(PRC)
	SYS_NPROC=getenv("SYS_NPROC");
# endif

	if (!MPI_NPROC) /* program not run from tmpirun */
		np=1; /* single threaded */
	else
		np=atoi(MPI_NPROC);

# if defined(PRC)
	if (!SYS_NPROC) 
		knp=np;
	else
		knp=atoi(SYS_NPROC);
# endif

  	thrMPI_Nproc=np;

	/* adjust root id */
	if (ROOT_ID > thrMPI_Nproc-1)
		ROOT_ID = thrMPI_Nproc-1;
	if (ROOT_ID < 0 )
		ROOT_ID = 0;

# if !defined(PRC)
	ukey_setid(ROOT_ID);
# else
	ROOT_ID=np-1; /* last one */
# endif

  	if (np<=0) {
    		fprintf(stderr,"Illegal parameter!\n");
    		thrMPI_exit(-1);
  	}

	anl_shmem_ginit();
	anl_shmem_uinit(ROOT_ID, GLOBAL_SHARE);

	/**
	 * Note after anl_shmem_ginit(), should use uexit instead of thrMPI_exit().
	 */

#ifdef MPISIM
	sim_global_init(np);
#endif

# if defined(PRC)
	mytp_init(); /* initialization */
# endif

	mb_global_init(np); /* buffer management */
	if (thrMPI_GCH_ginit()) {
		fprintf(stderr, "channel initialization failed"); /* channel management */
		uexit(-1);
	}

	twait_ginit();

	thrMPI_COMM_WORLD = (thrMPI_Comm**) anl_malloc (ROOT_ID, sizeof(thrMPI_Comm*)*np);
 	MPI_COMM_WORLD_BC=thrMPI_bco_create(np);
	MPI_COMM_WORLD_RDC=thrMPI_rdco_create(np);
	MPI_COMM_WORLD_ARDC=thrMPI_ardco_create(np);
	if (MPI_COMM_WORLD_BC==NULL || MPI_COMM_WORLD_RDC==NULL || MPI_COMM_WORLD_ARDC==NULL) {
		fprintf(stderr, "communicator initialization failed\n");
		uexit(-1);
	}

	MPI_COMM_WORLD_BAR=(thrMPI_Bar *)anl_malloc(ROOT_ID, sizeof(thrMPI_Bar));
	bar_init(MPI_COMM_WORLD_BAR, np);

	/* generated by mpitrans */
	thrMainInit();

  	Threads = (ukey_thr_t*) anl_malloc (ROOT_ID, sizeof(ukey_thr_t)*(thrMPI_Nproc-1));

	j=0;
  	for (i=0; i<=thrMPI_Nproc-1; i++) {
		if (i!=ROOT_ID) {
    			Threads[j]=ukey_thrcreate(thrMPI_Startup, i);
			j++;
		}
  	}

# if !defined(PRC)
	thrMPI_Startup(); /* remove it */
# else
	ukey_thrcreate(thrMPI_Startup, ROOT_ID);

	ukey_thrrun(knp);
# endif

  	for (i=0; i<thrMPI_Nproc-1; i++) {
		ukey_thrwait(Threads[i]);
  	}

#ifdef MPISIM
	sim_analyze(thrMPI_Nproc);
#endif

	anl_shmem_uend(ROOT_ID);
	anl_shmem_gend();

	return 0;
}

/* prototype, generated by mpitrans */
void thrUserInit(void);

static int thrMPI_UInit(int myid) 
{
	if (myid!=ROOT_ID)
		anl_shmem_uinit(myid, GLOBAL_SHARE);
	mb_init(myid, SYS_BUFFER);
  	comm_init(myid);
#ifdef MPISIM
	sim_init(myid);
#endif
	if (thrMPI_GCH_uinit(myid)) {
		fprintf(stderr, "channel user initialization failed\n");
	}

	twait_uinit(myid);
	
#ifdef MPISIM
	sim_gstart(myid);
#endif

	/* generated by mpitrans */
	thrUserInit();
	return MPI_SUCCESS;
}

static void thrMPI_UEnd(int myid) 
{
	if (myid!=ROOT_ID) {
		anl_shmem_uend(myid);
	}

	twait_uend(myid);
#ifdef MPISIM
	sim_gend(myid);
#endif
}

static void thrMPI_Startup(void) {
	int myid=ukey_getid();

	/** fprintf(stderr, "Thread %d starts to run\n", myid); **/
  	thrMPI_UInit(myid);
  	thrMPI_Barrier(MPI_COMM_WORLD);
  	tmain(ext_argc, ext_argv); /* entry point of the real MPI program */
  	thrMPI_UEnd(myid);
}

/* wrapping functions are necessary because ukey lib is just source */
/* level uniformed, to make it binary level uniformed, I have to make up */
/* these wrapping functions */
int thrMPI_createkey(void)
{
	ukey_t key;

	ukey_createkey(&key);

	/* don't be astonished, type of key must be simple */
	/* type that allows direct assignment */

	return (int)key;
}

void thrMPI_setval(int key, void *val)
{
	ukey_setval((ukey_t)key, val);
	return;
}

int thrMPI_getid()
{
	return ukey_getid();
}

void *thrMPI_getval(int key)
{
	return ukey_getval((ukey_t)key);
}
