/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

# define _REENTRANT

# include <stdio.h>
# include <stdlib.h>

# include "bcast.h"
# include "ukey.h"
# include "anl_shmem.h"
# include "twait.h"
# include "root.h"
/* # include "mbuffer.h" */

#ifdef MPISIM
#include "sim.h"
#endif

# define cm_alloc(b) malloc(b)
# define cm_free(b) free(b)

# define min(a,b) (((a)<(b))?(a):(b))

/* a lock free allocator because their could be other proc deallocate handles */
/* allocate a BCHandle for bco_array[which] */
static thrMPI_BCHandle *thrMPI_bch_alloc(thrMPI_BcastObj *bco, int which)
{
	register thrMPI_BCHandle **pfree;
	thrMPI_BCHandle *ret;
	register thrMPI_BCHandle *localfree;

	if (bco->group_size<2) return NULL; /* a group of one proc does not need handle */

	pfree=&(bco->bco_array[which]->free);
	ret=*pfree;

	while (ret) {
		localfree=ret->next;
		if (__compare_and_swap((long *)pfree, (long)ret, (long)localfree))
			break;
		ret=*pfree;
	}
		
	if (!ret) { /* no more cached handles */
		int i;

		ret=localfree=(thrMPI_BCHandle *)
			anl_malloc(which, sizeof(thrMPI_BCHandle)*bco->group_size);
		if (!localfree)
			return NULL;
		localfree++;

		/* link the chunk of handles */
		for (i=0; i<bco->group_size-2; i++)
			localfree[i].next=&(localfree[i+1]);
		
		/* now localfree[i] is the last one */
		do {
			localfree[i].next=*pfree;
		}
		while (!__compare_and_swap((long *)pfree, (long)localfree[i].next, (long)localfree));
	}	

	return ret;
}

/* a lock free deallocator */
static void thrMPI_bch_free(thrMPI_BcastObj *bco, int which, thrMPI_BCHandle *handle)
{
	register thrMPI_BCHandle **pfree=&(bco->bco_array[which]->free);

	do {
		handle->next=*pfree;
	} while (!__compare_and_swap((long *)pfree, (long)handle->next, (long)handle));

	return;
}

static thrMPI_BCQueue *thrMPI_bcq_create(int groupsize)
{
	thrMPI_BCQueue *ret;
	int i;

	if (groupsize<2) return NULL; /* a group of one proc does not need BCQueue */

	ret=(thrMPI_BCQueue *)
		anl_malloc(ROOT_ID, sizeof(thrMPI_BCQueue)+sizeof(thrMPI_BCHandle)*(groupsize-1));

	if (!ret) return NULL;
	if (ukey_createkey(&(ret->key_curhandle))) { /* key creation failed */
		anl_free(ROOT_ID, ret);
		return NULL;
	}

	ret->first=ret->cache;
	ret->free=ret->cache+1;

	for (i=0; i<groupsize-2; i++)
		ret->free[i].next=&(ret->free[i+1]);

	ret->free[i].next=NULL;

	ret->first->next=NULL;
	ret->first->avail=0; /* guarding node */

	return ret;
}
	
thrMPI_BcastObj *thrMPI_bco_create(int groupsize)
{
	thrMPI_BcastObj *ret;
	int i,j;

	ret=(thrMPI_BcastObj *)
		anl_malloc(ROOT_ID, sizeof(thrMPI_BcastObj)+sizeof(thrMPI_BCQueue *)*(groupsize-1));

	if (!ret) return NULL;

	ret->group_size=groupsize;
	
	if (groupsize<2) {
		ret->bco_array[0]=NULL;
		return ret;
	}

	for (i=0;i<groupsize;i++) {
		if ((ret->bco_array[i]=thrMPI_bcq_create(groupsize))==NULL) {
			for (j=0;j<i;j++)
				anl_free(ROOT_ID, ret->bco_array[j]);
			anl_free(ROOT_ID, ret);
			return NULL;
		}
	}

	return ret;
}

int thrMPI_bco_bcast(thrMPI_BcastObj *bco, int myid, int root, char *data, unsigned int len
#ifdef MPISIM
, int lseq
#endif
) {
	/** int myid; **/
	register thrMPI_BCHandle *curitem;
	thrMPI_BCHandle *newitem;
	register thrMPI_BCQueue *q;
	int ret;

	if (bco->group_size<2) return 0; /* don't do anything */ 

	q=bco->bco_array[root];

	curitem=(thrMPI_BCHandle *)ukey_getval(q->key_curhandle);
	if (!curitem)
		curitem=q->first;

	if (curitem==(thrMPI_BCHandle*)8L) /* internal allocation error */
		return -1;

	/* now curitem holds a handle whose avail is not -1 */

	/**
	fprintf(stderr, "%d in broadcast\n", myid);
	fflush(stderr);
	**/

	if (myid==root) { /* now avail must be zero, functioning like a lock */
		/* expend the queue by one */
		newitem=thrMPI_bch_alloc(bco, root);
		if (!newitem) {
# if _DEBUG_
			fprintf(stderr, "bcast failed due to insufficient memory\n");	
# endif	
			newitem=(thrMPI_BCHandle*)8L;
		}
		else {
			newitem->avail=0;
			newitem->next=NULL;
		}

		curitem->next=newitem;

# if (_DEBUG_)
		if (curitem->avail) { /* impossible */
			fprintf(stderr, "avail==%d when root comes\n", curitem->avail);
			return -1; 
		}
# endif

#ifdef MPISIM
		curitem->lseq=lseq;
#endif
		/* if ((curitem->data=anl_malloc(len))==NULL) { */
		/* if ((curitem->data=mb_alloc(myid, len))==NULL) { */
		if ((curitem->data=cm_alloc(len))==NULL) { 
# if (_DEBUG_)
			fprintf(stderr, "bcast failed due to insufficient memory\n");	
# endif	
			curitem->len=0;
			ret=-2; /* no memory */
		}
		else {
			memcpy(curitem->data, data, len);
			curitem->len=len;
			ret=0;
		}

		curitem->avail=1; /* the buffer is now available */

		/* same as curitem->next, but local data will ensure faster access */
		ukey_setval(q->key_curhandle, newitem);

 		/* better signal everybody, but don't do it now for easy debugging */
		/* signalAllFlag(&(curitem->avail), 1); */
		/* oh no, signalAll is too costly, why not just let recver wake by itself */
	}
	else { /* receivers */
		while (!curitem->avail) 
			waitFlagNE(&(curitem->avail), 0, root);

#ifdef MPISIM
		simMPI[root].event[curitem->lseq].children[__add_and_fetch(&(simMPI[root].event[curitem->lseq].child_num), 1)-1] = simMPI[myid].event[lseq].seq;
#endif

		if (curitem->data) {
			memcpy(data, curitem->data, min(len, curitem->len));
			ret=0;
		}
		else
			ret=-2;

		/* next item must be non-zero now */
		ukey_setval(q->key_curhandle, curitem->next);

		if (__add_and_fetch(&(curitem->avail), 1)==bco->group_size) { /* last to leave */
			if (!ret)
				cm_free(curitem->data); 
				/** mb_free(myid, curitem->data);  */
				/** anl_free(curitem->data); **/
			q->first=curitem->next;
			thrMPI_bch_free(bco, root, curitem);
		}
	}

	/**
	fprintf(stderr, "%d out of broadcast\n", myid);
	fflush(stderr);
	**/

	return ret;
}
