/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

/*
 * comm.c -- operations on communicator
 */

# include "struct.h"
# include "ukey.h"
#include "comm.h"

#define MPI_COMM_WORLD_PT2PT_CONTEXT 0
#define MPI_COMM_WORLD_COLL_CONTEXT 1

extern int thrMPI_Nproc;
extern thrMPI_Comm **thrMPI_COMM_WORLD;
extern thrMPI_Bar *MPI_COMM_WORLD_BAR;
extern thrMPI_BcastObj *MPI_COMM_WORLD_BC;
extern thrMPI_ReduceObj *MPI_COMM_WORLD_RDC;
extern thrMPI_AllreduceObj *MPI_COMM_WORLD_ARDC;

/* initialize the MPI_COMM_WORLD communicator */
int comm_init(int myid) {
  int i;
  thrMPI_COMM_WORLD[myid] = (thrMPI_Comm*) anl_malloc(myid, sizeof(thrMPI_Comm));
  thrMPI_COMM_WORLD[myid]->cookie = MPI_COMM_COOKIE;
  thrMPI_COMM_WORLD[myid]->np = thrMPI_Nproc;
  thrMPI_COMM_WORLD[myid]->local_rank=myid;
  thrMPI_COMM_WORLD[myid]->lrank2grank = (int*) anl_malloc(myid, thrMPI_Nproc*sizeof(int));
  for (i=0; i<thrMPI_Nproc; i++)
    thrMPI_COMM_WORLD[myid]->lrank2grank[i] = i;
  thrMPI_COMM_WORLD[myid]->pt2pt_context = MPI_COMM_WORLD_PT2PT_CONTEXT;
  thrMPI_COMM_WORLD[myid]->coll_context = MPI_COMM_WORLD_COLL_CONTEXT;
  thrMPI_COMM_WORLD[myid]->comm_type = INTRA;
  thrMPI_COMM_WORLD[myid]->group = (thrMPI_Group*) anl_malloc(myid, sizeof(thrMPI_Group));
  thrMPI_COMM_WORLD[myid]->group->cookie = MPI_GROUP_COOKIE;
  thrMPI_COMM_WORLD[myid]->group->np = thrMPI_Nproc;
  thrMPI_COMM_WORLD[myid]->group->local_rank = myid;
  thrMPI_COMM_WORLD[myid]->group->ref_count = 1;
  thrMPI_COMM_WORLD[myid]->group->permanent = True;
  thrMPI_COMM_WORLD[myid]->group->lrank2grank = thrMPI_COMM_WORLD[myid]->lrank2grank;
  thrMPI_COMM_WORLD[myid]->ref_count = 1;
  thrMPI_COMM_WORLD[myid]->bc_ptr = MPI_COMM_WORLD_BC;
  thrMPI_COMM_WORLD[myid]->rdc_ptr = MPI_COMM_WORLD_RDC;
  thrMPI_COMM_WORLD[myid]->ardc_ptr = MPI_COMM_WORLD_ARDC;
  thrMPI_COMM_WORLD[myid]->comm_cache = NULL;
  thrMPI_COMM_WORLD[myid]->attr_cache = NULL;
  thrMPI_COMM_WORLD[myid]->permanent = True;
  
  thrMPI_COMM_WORLD[myid]->br = MPI_COMM_WORLD_BAR;

  return MPI_SUCCESS;
}

#if 0
/* verify a communicator */
thrMPI_Comm* comm_verify(MPI_Comm comm, int myid) {
  	if (comm==MPI_COMM_WORLD) {
	/* myid=ukey_getid(); */
    		return thrMPI_COMM_WORLD[myid];
  	}

  	if (comm->cookie==MPI_COMM_COOKIE) 
		return comm;

  	return NULL;
}
#endif

/* get the number of processes in the communicator */
int MPI_Comm_size(MPI_Comm communicator, int *size) {
  	thrMPI_Comm* comm;
	int myid=ukey_getid();

  	if ((comm=comm_verify(communicator, myid))==NULL) 
		return MPI_ERR_COMM;

  	*size=comm->np;
	
  	return MPI_SUCCESS;
}

/* get the rank of calling process in this communicator */
int MPI_Comm_rank(MPI_Comm communicator, int *rank) {
  	thrMPI_Comm *comm;
	int myid=ukey_getid();

  	if ((comm=comm_verify(communicator, myid))==NULL) 
		return MPI_ERR_COMM;

  	*rank=comm->local_rank;

  	return MPI_SUCCESS;
}

/* compare two MPI communicators */
int MPI_Comm_compare(MPI_Comm communicator1, MPI_Comm communicator2, int *result) {
  	thrMPI_Comm *comm1, *comm2;
  	int group_result, ret_code;
	int myid=ukey_getid();

  	if ((comm1=comm_verify(communicator1, myid))==NULL) 	
		return MPI_ERR_COMM;
  	if ((comm2=comm_verify(communicator2, myid))==NULL) 
		return MPI_ERR_COMM;

  	MPI_Group_compare(comm1->group, comm2->group, &group_result);
  	switch (group_result) {
  	case MPI_IDENT:
    		if (comm1->pt2pt_context==comm2->pt2pt_context && \
		    comm1->coll_context==comm2->coll_context)
      			*result=MPI_IDENT;
    		else
      			*result=MPI_CONGRUENT;
    		break;
  	case MPI_SIMILAR:
    		*result=MPI_SIMILAR;
    		break;
  	case MPI_UNEQUAL:
    		*result=MPI_UNEQUAL;
    		break;
  	}

  	return MPI_SUCCESS;
}
