/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

#ifndef _struct_h_
#define _struct_h_

#define _REENTRANT

#include <sys/types.h>
#include <errno.h>
#include <limits.h>
#include <unistd.h>
#include <stdio.h>

#include "mpi.h"

typedef unsigned char Boolean;
enum {False=0, True=1};
typedef unsigned char thrMPI_Mode; 
enum {STANDARD=1, READY=2, SYNC=4, BUFFERED=8}; 
typedef enum {INTER, INTRA} thrMPI_Comm_type;

/* #define DEBUG */

# include "sync.h"

/***********************************************************************/
/* a group of processes                                                */
/***********************************************************************/

#define MPI_GROUP_COOKIE 0x87ef54a0

typedef struct thrMPI_Group {
  unsigned long cookie;	/* cookie to help detect valid item */
  int np;		/* number of processes in group */
  int local_rank;	/* my rank in the group (if I belong) */
			/* MPI_UNDEFINED if I am not a member */
  int ref_count;	/* number of references to this group */
  Boolean permanent;	/* permanent group */
  int *lrank2grank;	/* mapping from local to "global" ranks */  
} thrMPI_Group;


/***********************************************************************/
/* Communicator is a group of processes involved in communication      */
/***********************************************************************/

# include "barrier.h"
# include "bcast.h"
# include "reduce.h"

#define MPI_COMM_COOKIE 0xed56780a

typedef struct thrMPI_Comm {
  unsigned long cookie;		/* cookie to help detect valid item */
  int np;			/* size of group */
  int local_rank;		/* rank in local_group for this process */
  int *lrank2grank;		/* mapping for group */
  unsigned long pt2pt_context;	/* context for point to point communication */
  unsigned long coll_context;	/* context for collective communication */
  thrMPI_Comm_type comm_type;	/* inter or intra */
  thrMPI_Group *group;		/* group associated with this communicator */
  int ref_count;		/* number of references to communicator */

  thrMPI_BcastObj *bc_ptr;	/* bcast object, hook for broadcast */
  thrMPI_ReduceObj *rdc_ptr;	/* reduce object, hook for reduce */
  thrMPI_AllreduceObj *ardc_ptr;/* allreduce object, hook for allreduce */

  void *comm_cache;		/* hook for communicator cache */
  void *attr_cache;		/* hook for attribute cache */
  Boolean permanent;		/* permanent communicator */

  thrMPI_Bar *br;		/* barrier structure, one copy shared by all */
} thrMPI_Comm;


/***********************************************************************/
/* Data structure for send/receive handle and request                  */
/***********************************************************************/

#define MPI_REQUEST_COOKIE 0xce45b310

enum {
  MPI_SEND=1, 
  MPI_RECV=2, 
  MPI_PERSISTENT_SEND=4,
  MPI_PERSISTENT_RECV=8,
  MPI_IMMEDIATE_SEND=16,
  MPI_IMMEDIATE_RECV=32
}; 

typedef unsigned char thrMPI_OPTYPE;

/* states of message handles */
enum {thrMPI_CREATE=0x1, thrMPI_PEND=0x2, thrMPI_MATCH=0x4, thrMPI_FREE=0x8, \
	thrMPI_CANCEL=0x10, thrMPI_AVAIL=0x20, thrMPI_DONE=0x40, thrMPI_ERR=0x80};

typedef struct thrMPI_COMMON {
  thrMPI_OPTYPE handle_type;
  unsigned long cookie;
  unsigned int stat; 		/* handle state */
  unsigned long context;	/* context id */
  int tag;			/* tag */
  int src;			/* global source process for message */
  int dest;			/* global destination process for message */
# if (!_LOCK_)
  int my_ts;
  int peer_ts;
# endif
} thrMPI_Common;

/* message send handle */
typedef struct thrMPI_Shandle {
  thrMPI_OPTYPE handle_type;
  unsigned long cookie;		/* Cookie to help detect valid item */
  unsigned int stat; 		/* handle state */
  unsigned long context;	/* context id */
  int tag;			/* tag */
  int src;			/* global source process for message */
  int dest;			/* global destination process for message */
# if (!_LOCK_)
  int my_ts;
  int peer_ts;
# endif
# if (_DEBUG_)
  MPI_Datatype datatype;	/* basic or derived datatype */
# endif
  int count;			/* requested count */
  MPI_Comm comm;		/* communicator */
  char *bufadd;			/* address of buffer, can be in-place buffer or sysbuf */
  char *user_buf;		/* when useSysbuf==True, bufadd is the address of */
				/* system buffer and user_buf the starting address */
				/* of user data */
  int buflen;			/* length of buffer at bufadd */
#ifdef MPISIM
  int lseq;	/* sequence number as simulation event */
#endif
  thrMPI_Mode mode;		/* mode: STANDARD, READY, SYNC, BUFFERED */
  Boolean useSysBuf;		/* use system buffer or not */
  /* Boolean persistent; */		/* persistent operation or not */
  Boolean active;			/* relavent only persistent handle */
  int errval;			/* Holds any error code; 0 for none */
#if (_LOCK_)
  int I_have_no_interest;
#endif

# if (_SR_CHECK_)
  int seq_no;
# endif

} thrMPI_SHandle;

/* message receive handle */
typedef struct thrMPI_Rhandle {
  thrMPI_OPTYPE handle_type;
  unsigned long cookie;		/* Cookie to help detect valid item */
  unsigned int stat; 		/* handle state */
  unsigned long context;	/* context id */
  int tag;			/* tag */
  int src;			/* global source process for message */
  int dest;			/* global destination process for message */
#if (!_LOCK_)
  int my_ts;
  int peer_ts;
#endif
  MPI_Datatype datatype;	/* basic or derived datatype */
  int count;			/* requested count */
  MPI_Comm comm;		/* communicator */
  char *bufadd;			/* address of buffer */
  int buflen;			/* length of buffer at bufadd */
#ifdef MPISIM
  int lseq;	/* sequence number as simulation event */
#endif
  /* thrMPI_Status *status; */ /* not necessary */
  int errval;			/* Holds any error code; 0 for none */
  /* Boolean persistent; */		/* persistent operation or not */
  Boolean active;			/* relavent only persistent handle */
  /* For persistant receives, we may need to restore some of the fields
     after the operation completes */
  /* the parameters set at the construction time of the request */
  int perm_source, perm_tag, perm_len;

# if (_SR_CHECK_)
  int seq_no;
# endif

} thrMPI_RHandle;
	
union thrMPI_Handle {
  thrMPI_OPTYPE type;    
  thrMPI_SHandle shandle;
  thrMPI_RHandle rhandle;
  thrMPI_SHandle persistent_shandle;
  thrMPI_RHandle persistent_rhandle;
  thrMPI_Common common;
};

#define MIN(a,b) (((a)<(b))? (a):(b))
#define MAX(a,b) (((a)>(b))? (a):(b))

# include "twait.h"

#endif
