/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

# ifndef _TQUEUE_H_
# define _TQUEUE_H_

#define _REENTRANT

/* interface: 	Q_create, Q_init, Q_reset, enQ, deQ, Q_probe, Q_probe_last, 
		Q_find, Q_free, Q_removeElem, Q_newElem, Q_freeElem */
/* Q_create implies Q_init, cmp function can be NULL, in which case the queue 
 * is just a standard queue. Queues created by Q_create must be freed by Q_free.
 * Q_init initialize the queue, this is useful when we already have a TQueue object
 * e.g. in stack or inside another structure.
 * Q_reset resets the queue to intial states, all elements in the queue are freed 
 *
 * enQ add one element to the end of the queue
 * deQ move the element at the head of the queue, if not empty, from the queue
 *
 * Q_probe, Q_find are not enabled until a non-NULL cmp function is provided for Q_create, 
 * Q_probe and Q_find extend the queue to a simple priority queue. The cmp takes an element 
 * and a context-tag. The tag can be interpreted as a priority threshold or just a TAG.  
 * Both Q_probe and Q_find return the first element searching from the head of the queue such 
 * that cmp(elem, tag) returns 0, or NULL if none of such elements exist. Q_probe differs from
 * Q_find in that Q_probe does not remove the element from the queue while Q_find does.
 * In other words, Q_probe + Q_removeElem == Q_find.
 *
 * Q_probe_last requires one more argument as the element returned from the last time 
 * Q_probe/Q_probe_last is called. User must ensure the element is still in the queue or 
 * there could be disastrous results.
 * 
 * All elements are to be allocated/deallocated by Q_newElem/Q_freeElem
 */

/* Mater's traverse function: Q_first, Q_last, Q_next. They can only be called by
 * the master. But the library does not enforce the check.
 */

/* Q_first returns the first item in the queue, if queue empty , returns NULL;
 *
 * Q_next takes an item as parameter and returns the following item in the queue, 
 * if already at the end of the queue, return NULL;
 *
 * Q_last returns the end of the queue, if queue empty, return NULL.
 */ 

/* interface augmentation: Q_safe, Q_incr, Q_decr, Q_ucnt, Q_resetUcnt */
/* These new functions differ from the above in that they are not lock free or they
 * need some synchronization primitives such as compare_and_swap, add_and_fetch, etc.
 * They are provided specifically for the MPI routines.
 */

/* Q_safe will block until the item is safe to be reused;
 * Q_incr will atomically add a specific number to a user counter and return the value
 * of the counter after the addition. 
 * 
 * Q_decr, similarly, will atomically subtract a number from a user counter and return the
 * value of the counter after the subtraction. Note, Q_decr(q, n) is equivalent to 
 * Q_incr(q, -n).
 *
 * Q_ucnt returns the current value of the count. It is equivalent to Q_incr(q, 0).
 *
 * Q_resetUcnt will reset the count to be 0. But at the return time, it is not guarantteed
 * the value of the user counter still be 0.
 */

typedef struct BUCKET
{
	struct BUCKET *next;
	struct BUCKET *prev;
} BUCKET;

/* The concurrency model of this queue implementation is not quite general, 
 * it always assumes only one process can add/remove elements in the queue, 
 * which is called the master of the queue.
 * A one-way channel from proc i to proc j consists two queues, sq[i][j] and rq[i][j]. 
 * proc i is the master of sq[i][j] and proc j is the master of rq[i][j].
 */
typedef struct queue
{
	BUCKET 	qctrl;	/* next to the head and prev to the end of the queue */
	int	nts; /* next available time stamp value */
	int 	count;	/* number of elements currently in queue */
	int 	ucount;	/* not used by queue management, but can be used by user */
	int 	(*cmp)(void *item1, int context, int tag, int arg
#if (!_LOCK_)
, void *arg2
#endif
);/* comparison function */
	/* those elements in the freed list will be deferred when nTraverse==0 */
	BUCKET * freed; /* linked list of freed elements with BUCKET's prev link, */ 
			/* whose next field may be in use */
	int 	nTraverse; /* number of processes traversing the queue */

	BUCKET * allocate; /* linked list of allocable BUCKET's */
	int 	owner;
	int	item_size; /* items in each queue are equal sized */
	int 	chunk_group; /* queue itemsare allocated in chunks */
			     /* since handles are recycled, chunk_group */
			     /* is not quite sensitive to performance */
}TQueue;

# define QHEAD(q) ((q)->qctrl.next)
# define QTAIL(q) ((q)->qctrl.prev)

extern void *Q_removeElem(TQueue *queue, void *elem);
extern void *deQ(TQueue *queue);

# if (!_LOCK_)
/* the three functions may be efficient if used correctly but could */
/* cause problems if used incorrectly, e.g. forget to call probe_last_finish */
/* It is recommeded to program first without them, and after everything works */
/* use them to substitute the probe calls if applicable. */
extern void Q_probe_last_init(TQueue *queue);
extern void *Q_probe_last(TQueue *queue, int context, int tag, int arg, void *arg2, void *elem);
extern void Q_probe_last_finish(TQueue *queue);

extern int Q_isSafe(TQueue *queue);
extern int Q_isEmpty(TQueue *queue);
extern int Q_now(TQueue *queue); /* return the current timestamp */

/* MPI uses the ucnt to reflect the number of free handles in the queue */
/* A scan by the master of the queue will be triggered when the number of free handles 
 * is large enough to hurt the performance. (E.g. (ucnt>=5 && (ncnt/count)>threshold) )
 */
extern int Q_incr(TQueue *queue, int n);
extern int Q_decr(TQueue *queue, int n);
extern int Q_ucnt(TQueue *queue);
extern void Q_resetUcnt(TQueue *queue);
extern int Q_cnt(TQueue *queue); /* return the length of the queue */
# endif

extern void *Q_newElem(TQueue *queue);
extern void Q_freeElem(TQueue *queue, void *elem);
extern TQueue * Q_create( int (* cmp_function)(void *elem, int context, int tag, int arg
#if (!_LOCK_)
, void *arg2
#endif
), \
	int item_size, int chunk_group, int owner);
extern void Q_init(TQueue *queue, int (* cmp_function)(void *elem, int context, int tag, int arg
#if (!_LOCK_)
, void *arg2
#endif
), \
	int item_size, int chunk_group, int owner); /* init the queue */
# if (!_LOCK_)
extern void Q_reset(TQueue *queue); /* set queue to be an empty queue */
# endif
extern void enQ(TQueue *queue, void *elem);
extern int Q_safe(TQueue *queue, void *elem);

#if (_LOCK_)
extern void *Q_probe(TQueue *queue, int context, int tag, int arg);
#else
extern void *Q_probe(TQueue *queue, int context, int tag, int arg, void *arg2);
#endif

extern void *Q_find(TQueue *queue, int context, int tag, int arg, void *arg2);
extern void *Q_find_last(TQueue *queue, int context, int tag, int arg, void *arg2, void *elem);
extern void Q_free(TQueue *queue);

# if (!_LOCK_)
extern void *Q_first(TQueue *queue);
extern void *Q_next(TQueue *queue, void *item);
extern void *Q_last(TQueue *queue);
# endif

# endif
