/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

# ifndef _UKEY_H_
# define _UKEY_H_

#define _REENTRANT

/**
 * This file defines a uniformed interface of thread specific data (TSD) 
 * mechanism. Currently the following thread systems are supported:
 * sproc(shared address space) and pthread. Note, no key deallocation function 
 * is provided because the key allocation/deallocation needs to be serialized 
 * and it is recommended to create keys at the initial phase of the program. 
 * It is not recommended to create a process (of separate address space) from a
 * thread which keeps an identical copy of address space and also uses ukey 
 * functions. (But it is OK to fork a process and then exec some external 
 * program or doesn't use any ukey functions after the fork.) The result has
 * not been carefully tested.
 *
 * ukey_thr_t ukey_thrcreate(start_routine, id) creates a thread with internal
 *	id. Id is a short integer. Note: the system's maximum number of 
 *	processes in a shared group is 10000.
 *
 * int ukey_thrwait(thread) waits until the thread finishes execution.
 *
 * void ukey_setid(id) is provided so that old programs that directly calls
 * 	pthread_create/sproc can be modified by adding this call to the top
 * 	of its start routine. This is not recommended because id should 
 * 	be associated with the thread by the time it is created. (Note: 
 *	ukey_thrcreate is equivalent to a call to pthread_create/sproc, then 
 *	ukey_setid, then start_routine.
 *
 * short ukey_getid() returns the internal id associated with the calling 
 *	thread.	It will return -1 if the thread is not created by 
 *	ukey_thrcreate or no ukey_setid is called before.
 *
 * int ukey_createkey(pkey) creates a new key which is retured to the key
 *	structure pointed by pkey. 
 *
 * int ukey_setval(key, val) associates val to key. Val must be of type 
 *	(void *). By default, keys are associated with value 0.
 *
 * void * ukey_getval(key) returns the associated val of key. 
 *
 * void ukey_setkeymax(max) sets the maximum number of keys. Default maximum 
 * 	is 128. Only effective for sproc version currently.
 */

# include "thrsys.h"

# if (THREAD_SYS==PTHREAD_SYS)

# include <stdio.h>
# include <pthread.h>
typedef pthread_t * ukey_thr_t;
typedef pthread_key_t ukey_t;
# define ukey_thrcreate pthread_ukey_thrcreate
# define ukey_setid pthread_ukey_setid
# define ukey_getid pthread_ukey_getid
# define ukey_createkey(pkey) pthread_key_create(pkey, NULL)
# define ukey_setval(key, val) pthread_setspecific(key, val)
# define ukey_getval(key) pthread_getspecific(key)
# define ukey_setkeymax(max) fprintf(stderr, "ukey_setkeymax is not defined for pthread system!\n")
# define ukey_thrwait(thr) ( (pthread_join((*thr), NULL)) ? (free(thr), -1) : (free(thr), 0) )

# else

# include <sys/types.h>
typedef pid_t ukey_thr_t;
typedef int ukey_t;
# define ukey_thrcreate sproc_ukey_thrcreate
# define ukey_setid sproc_ukey_setid
# define ukey_getid sproc_ukey_getid
# define ukey_createkey sproc_ukey_createkey
# define ukey_setval sproc_ukey_setval
# define ukey_getval sproc_ukey_getval
# define ukey_thrwait(thr) waitpid(thr, NULL, NULL)

int sproc_ukey_createkey(ukey_t *pkey);
int sproc_ukey_setval(ukey_t key, const void *val);
void * sproc_ukey_getval(ukey_t key);
void ukey_setkeymax(int max);
# endif

ukey_thr_t ukey_thrcreate(void (*start_routine)(void), short id);
int ukey_setid(short id);
short ukey_getid(void);

# endif
