/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

/*
 * Send/Recv family of functions.
 * [I][bsr]send, [bsr]send_init, [I]recv, recv_init. 15 functions.
 * Sendrecv[_replace], two functions.
 */
# define _REENTRANT
# include <stdio.h>
# include "struct.h"

# include "mqueue.h"
# include "mpi.h"

# if (_LOCK_)
# include "lock.h"
# endif

#include "sim.h"

# if (_DEBUG_)
static void sh_set_datatype(void *sh, MPI_Datatype datatype)
{
	((thrMPI_SHandle *)sh)->datatype=datatype;
}
# endif

int thrMPI_Send(
	void *message,
	int count,
	MPI_Datatype datatype,
	int dest,
	int tag,
	MPI_Comm comm)
{
	MPI_Request req;
	int buflen=thrMPI_Usize(datatype)*count;
	int myid=ukey_getid();
	int ret;
#ifdef MPISIM
	int lseq=sim_comm_start(myid, Event_Send, comm, 0, buflen);
#endif

	ret=thrMPI_SH_init(comm, (char *)message, buflen, myid, dest, tag, count, STANDARD, False, False, &req);
#ifdef MPISIM
	((thrMPI_SHandle*)req)->lseq = lseq;
#endif

# if (_DEBUG_)
	sh_set_datatype(req, datatype);
# endif

	if (ret!=MPI_SUCCESS) {
		thrMPI_Request_free((MPI_Request *)&req);
# if (_DEBUG_)
		fprintf(stderr, "sh_init failed\n");
# endif
		return ret;
	}
	
	ret=thrMPI_Start((MPI_Request *)&req);
	if (ret!=MPI_SUCCESS) {
# if (_DEBUG_)
		fprintf(stderr, "send, start failed\n");
# endif
		thrMPI_Request_free((MPI_Request *)&req);
		return ret;
	}

#ifdef MPISIM
	sim_comm_end(myid);
#endif
	return thrMPI_Wait((MPI_Request *)&req, NULL);
}

int thrMPI_Isend(
	void *message,
	int count,
	MPI_Datatype datatype,
	int dest,
	int tag,
	MPI_Comm comm,
	MPI_Request *request)
{
	MPI_Request req;
	int buflen;
	int myid;
	int ret;

	buflen=thrMPI_Usize(datatype)*count;
	myid=ukey_getid();

	ret=thrMPI_SH_init(comm, (char *)message, buflen, myid, dest, tag, count, STANDARD, False, True, &req);

# if (_DEBUG_)
	sh_set_datatype(req, datatype);
#endif

	if (ret!=MPI_SUCCESS) {
# if (_DEBUG_)
		fprintf(stderr, "sh_init failed\n");
# endif
		thrMPI_Request_free((MPI_Request *)&req);
		(*request)=MPI_REQUEST_NULL;
		return ret;
	}
	
	ret=thrMPI_Start((MPI_Request *)&req);

	/* do not do error checking here because most programmer prefer to check */
	/* error upon return from MPI_Wait */
	*request=req; 
	
	return ret;
}

int thrMPI_Send_init(
	void *message,
	int count,
	MPI_Datatype datatype,
	int dest,
	int tag,
	MPI_Comm comm,
	MPI_Request *request)
{
	MPI_Request req;
	int buflen;
	int myid;
	int ret;

	buflen=thrMPI_Usize(datatype)*count;
	myid=ukey_getid();

	ret=thrMPI_SH_init(comm, (char *)message, buflen, myid, dest, tag, count, STANDARD, True, False, &req);

# if (_DEBUG_)
	sh_set_datatype(req, datatype);
#endif

	if (ret!=MPI_SUCCESS) {
		thrMPI_Request_free((MPI_Request *)&req);
		(*request)=MPI_REQUEST_NULL;
	}
	else {
		(*request)=req;
	}
	
	return ret;
}

int thrMPI_Bsend(
	void *message,
	int count,
	MPI_Datatype datatype,
	int dest,
	int tag,
	MPI_Comm comm)
{
	MPI_Request req;
	int buflen=thrMPI_Usize(datatype)*count;
	int myid=ukey_getid();
	int ret;
#ifdef MPISIM
	int lseq=sim_comm_start(myid, Event_Send, comm, 0, buflen);
#endif

	ret=thrMPI_SH_init(comm, (char *)message, buflen, myid, dest, tag, count, BUFFERED, False, False, &req);
#ifdef MPISIM
	((thrMPI_SHandle*)req)->lseq = lseq;
#endif

# if (_DEBUG_)
	sh_set_datatype(req, datatype);
#endif

	if (ret!=MPI_SUCCESS) {
		thrMPI_Request_free((MPI_Request *)&req);
		return ret;
	}
	
	ret=thrMPI_Start((MPI_Request *)&req);
	if (ret!=MPI_SUCCESS) {
		thrMPI_Request_free((MPI_Request *)&req);
		return ret;
	}
#ifdef MPISIM
	sim_comm_end(myid);
#endif
	return thrMPI_Wait((MPI_Request *)&req, NULL);
}

int thrMPI_Ibsend(
	void *message,
	int count,
	MPI_Datatype datatype,
	int dest,
	int tag,
	MPI_Comm comm,
	MPI_Request *request)
{
	MPI_Request req;
	int buflen;
	int myid;
	int ret;

	buflen=thrMPI_Usize(datatype)*count;
	myid=ukey_getid();

	ret=thrMPI_SH_init(comm, (char *)message, buflen, myid, dest, tag, count, BUFFERED, False, True, &req);

# if (_DEBUG_)
	sh_set_datatype(req, datatype);
#endif

	if (ret!=MPI_SUCCESS) {
		thrMPI_Request_free((MPI_Request *)&req);
		(*request)=MPI_REQUEST_NULL;
		return ret;
	}
	
	ret=thrMPI_Start((MPI_Request *)&req);

	(*request)=req;
	
	return ret;
}

int thrMPI_Bsend_init(
	void *message,
	int count,
	MPI_Datatype datatype,
	int dest,
	int tag,
	MPI_Comm comm,
	MPI_Request *request)
{
	MPI_Request req;
	int buflen;
	int myid;
	int ret;

	buflen=thrMPI_Usize(datatype)*count;
	myid=ukey_getid();

	ret=thrMPI_SH_init(comm, (char *)message, buflen, myid, dest, tag, count, BUFFERED, True, False, &req);

# if (_DEBUG_)
	sh_set_datatype(req, datatype);
#endif

	if (ret!=MPI_SUCCESS) {
		thrMPI_Request_free((MPI_Request *)&req);
		(*request)=MPI_REQUEST_NULL;
	}
	else {
		(*request)=req;
	}
	
	return ret;
}

int thrMPI_Rsend(
	void *message,
	int count,
	MPI_Datatype datatype,
	int dest,
	int tag,
	MPI_Comm comm)
{
	MPI_Request req;
	int buflen;
	int myid;
	int ret;

	buflen=thrMPI_Usize(datatype)*count;
	myid=ukey_getid();

	ret=thrMPI_SH_init(comm, (char *)message, buflen, myid, dest, tag, count, READY, False, False, &req);

# if (_DEBUG_)
	sh_set_datatype(req, datatype);
#endif

	if (ret!=MPI_SUCCESS) {
		thrMPI_Request_free((MPI_Request *)&req);
		return ret;
	}
	
	ret=thrMPI_Start((MPI_Request *)&req);
	if (ret!=MPI_SUCCESS) {
		thrMPI_Request_free((MPI_Request *)&req);
		return ret;
	}
	return thrMPI_Wait((MPI_Request *)&req, NULL);
}

int thrMPI_Irsend(
	void *message,
	int count,
	MPI_Datatype datatype,
	int dest,
	int tag,
	MPI_Comm comm,
	MPI_Request *request)
{
	MPI_Request req;
	int buflen;
	int myid;
	int ret;

	buflen=thrMPI_Usize(datatype)*count;
	myid=ukey_getid();

	ret=thrMPI_SH_init(comm, (char *)message, buflen, myid, dest, tag, count, READY, False, True, &req);

# if (_DEBUG_)
	sh_set_datatype(req, datatype);
#endif

	if (ret!=MPI_SUCCESS) {
		thrMPI_Request_free((MPI_Request *)&req);
		(*request)=MPI_REQUEST_NULL;
		return ret;
	}
	
	ret=thrMPI_Start((MPI_Request *)&req);
	(*request)=req;
	
	return ret;
}

int thrMPI_Rsend_init(
	void *message,
	int count,
	MPI_Datatype datatype,
	int dest,
	int tag,
	MPI_Comm comm,
	MPI_Request *request)
{
	MPI_Request req;
	int buflen;
	int myid;
	int ret;

	buflen=thrMPI_Usize(datatype)*count;
	myid=ukey_getid();

	ret=thrMPI_SH_init(comm, (char *)message, buflen, myid, dest, tag, count, READY, True, False, &req);

# if (_DEBUG_)
	sh_set_datatype(req, datatype);
#endif

	if (ret!=MPI_SUCCESS) {
		thrMPI_Request_free((MPI_Request *)&req);
		(*request)=MPI_REQUEST_NULL;
	}
	else {
		(*request)=req;
	}
	
	return ret;
}

int thrMPI_Ssend(
	void *message,
	int count,
	MPI_Datatype datatype,
	int dest,
	int tag,
	MPI_Comm comm)
{
	MPI_Request req;
	int buflen;
	int myid;
	int ret;

	buflen=thrMPI_Usize(datatype)*count;
	myid=ukey_getid();

	ret=thrMPI_SH_init(comm, (char *)message, buflen, myid, dest, tag, count, SYNC, False, False, &req);

# if (_DEBUG_)
	sh_set_datatype(req, datatype);
#endif

	if (ret!=MPI_SUCCESS) {
		thrMPI_Request_free((MPI_Request *)&req);
		return ret;
	}
	
	ret=thrMPI_Start((MPI_Request *)&req);
	if (ret!=MPI_SUCCESS) {
		thrMPI_Request_free((MPI_Request *)&req);
		return ret;
	}
	return thrMPI_Wait((MPI_Request *)&req, NULL);
}

int thrMPI_Issend(
	void *message,
	int count,
	MPI_Datatype datatype,
	int dest,
	int tag,
	MPI_Comm comm,
	MPI_Request *request)
{
	MPI_Request req;
	int buflen;
	int myid;
	int ret;

	buflen=thrMPI_Usize(datatype)*count;
	myid=ukey_getid();

	ret=thrMPI_SH_init(comm, (char *)message, buflen, myid, dest, tag, count, SYNC, False, True, &req);

# if (_DEBUG_)
	sh_set_datatype(req, datatype);
#endif

	if (ret!=MPI_SUCCESS) {
		thrMPI_Request_free((MPI_Request *)&req);
		(*request)=MPI_REQUEST_NULL;
		return ret;
	}
	
	ret=thrMPI_Start((MPI_Request *)&req);
	(*request)=req;
	
	return ret;
}

int thrMPI_Ssend_init(
	void *message,
	int count,
	MPI_Datatype datatype,
	int dest,
	int tag,
	MPI_Comm comm,
	MPI_Request *request)
{
	MPI_Request req;
	int buflen;
	int myid;
	int ret;

	buflen=thrMPI_Usize(datatype)*count;
	myid=ukey_getid();

	ret=thrMPI_SH_init(comm, (char *)message, buflen, myid, dest, tag, count, SYNC, True, False, &req);

# if (_DEBUG_)
	sh_set_datatype(req, datatype);
#endif

	if (ret!=MPI_SUCCESS) {
		thrMPI_Request_free((MPI_Request *)&req);
		(*request)=MPI_REQUEST_NULL;
	}
	else {
		(*request)=req;
	}
	
	return ret;
}

int thrMPI_Recv(
	void *message,
	int count,
	MPI_Datatype datatype,
	int source,
	int tag,
	MPI_Comm comm,
	MPI_Status *status)
{
	MPI_Request req;
	int buflen;
	int myid=ukey_getid();
	int ret;
#ifdef MPISIM
	int lseq=sim_comm_start(myid, Event_Recv, comm, 0, 0);
#endif

	buflen=thrMPI_Usize(datatype)*count;

	ret=thrMPI_RH_init(comm, (char *)message, buflen, source, myid, tag, datatype, count, False, False, &req);
#ifdef MPISIM
	((thrMPI_RHandle*)req)->lseq = lseq;
#endif

	if (ret!=MPI_SUCCESS) {
# if (_DEBUG_)
		fprintf(stderr, "rh_init failed\n");
# endif
		thrMPI_Request_free((MPI_Request *)&req);
		return ret;
	}
	
	ret=thrMPI_Start((MPI_Request *)&req);
	if (ret!=MPI_SUCCESS) {
# if (_DEBUG_)
		fprintf(stderr, "recv start failed\n");
# endif
		thrMPI_Request_free((MPI_Request *)&req);
		return ret;
	}
	/* fprintf(stderr, "recv started\n"); */
	ret=thrMPI_Wait((MPI_Request *)&req, status);
	/* fprintf(stderr, "recv done\n");  */

#ifdef MPISIM
	sim_comm_end(myid);
#endif
	return ret;
}

int thrMPI_Irecv(
	void *message,
	int count,
	MPI_Datatype datatype,
	int source,
	int tag,
	MPI_Comm comm,
	MPI_Request *request)
{
	MPI_Request req;
	int buflen;
	int myid;
	int ret;

	buflen=thrMPI_Usize(datatype)*count;
	myid=ukey_getid();

	ret=thrMPI_RH_init(comm, (char *)message, buflen, source, myid, tag, datatype, count, False, True, &req);
	if (ret!=MPI_SUCCESS) {
# if (_DEBUG_)
		fprintf(stderr, "rh_init failed\n");
# endif
		thrMPI_Request_free((MPI_Request *)&req);
		(*request)=MPI_REQUEST_NULL;
		return ret;
	}
	
	ret=thrMPI_Start((MPI_Request *)&req);
	(*request)=req;
	
	return ret;
}

int thrMPI_Recv_init(
	void *message,
	int count,
	MPI_Datatype datatype,
	int source,
	int tag,
	MPI_Comm comm,
	MPI_Request *request)
{
	MPI_Request req;
	int buflen;
	int myid;
	int ret;

	buflen=thrMPI_Usize(datatype)*count;
	myid=ukey_getid();

	ret=thrMPI_RH_init(comm, (char *)message, buflen, source, myid, tag, datatype, count, True, False, &req);
	if (ret!=MPI_SUCCESS) {
		thrMPI_Request_free((MPI_Request *)&req);
		(*request)=MPI_REQUEST_NULL;
	}
	else {
		(*request)=req;
	}
	
	return ret;
}

int thrMPI_Sendrecv(
	void *send_buf,
	int send_count,
	MPI_Datatype send_type,
	int dest,
	int send_tag,
	void *recv_buf,
	int recv_count,
	MPI_Datatype recv_type,
	int source,
	int recv_tag,
	MPI_Comm comm,
	MPI_Status *status)
{
	MPI_Request req1, req2;
	int ret1, ret2;

/***
	ret1=MPI_Irsend(send_buf, send_count, send_type, dest, send_tag, comm, &req1);
	if (ret1!=MPI_SUCCESS) {
		MPI_Request_free(&req1);
		MPI_Isend(send_buf, send_count, send_type, dest, send_tag, comm, &req1);
	}
****/
	MPI_Isend(send_buf, send_count, send_type, dest, send_tag, comm, &req1);
	MPI_Irecv(recv_buf, recv_count, recv_type, source, recv_tag, comm, &req2);
	ret1=MPI_Wait(&req1, NULL);
	ret2=MPI_Wait(&req2, status);
	if (ret1!=MPI_SUCCESS)
		return ret1;
	if (ret2!=MPI_SUCCESS)
		return ret2;
	return MPI_SUCCESS;
}

int thrMPI_Sendrecv_replace(
	void *buffer,
	int count,
	MPI_Datatype datatype,
	int dest,
	int send_tag,
	int src,
	int recv_tag,
	MPI_Comm comm,
	MPI_Status *status)
{
	MPI_Request req1, req2;
	int ret1, ret2;
	void *recv_buf;
	
	ret1=MPI_Irsend(buffer, count, datatype, dest, send_tag, comm, &req1);
	if (ret1!=MPI_SUCCESS) { /* need extra buffering */
		MPI_Request_free(&req1);
		MPI_Ibsend(buffer, count, datatype, dest, send_tag, comm, &req1);
	}
	MPI_Irecv(buffer, count, datatype, src, recv_tag, comm, &req2);
	ret1=MPI_Wait(&req1, NULL);
	ret2=MPI_Wait(&req2, status);
	if (ret1!=MPI_SUCCESS)
		return ret1;
	if (ret2!=MPI_SUCCESS)
		return ret2;
	return MPI_SUCCESS;
}
