/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

#ifdef MPISIM

/* sim.c - routines for simulation purposes */

#include <time.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>

#include "root.h"
#include "anl_shmem.h"
#include "comm.h"
#include "sim.h"

simMPI_ctrl_t* simMPI;

/* global initialization */
void sim_global_init(int nproc) {
  simMPI = (simMPI_ctrl_t*) anl_malloc (ROOT_ID, sizeof(simMPI_ctrl_t)*nproc);
}

/* initialization by each individual MPI node */
void sim_init(int myid) {
  simMPI[myid].MAX_EVENT_NUM=100;
  simMPI[myid].event=(simMPI_Event*) anl_malloc (myid, simMPI[myid].MAX_EVENT_NUM*sizeof(simMPI_Event));
  simMPI[myid].EventNum=0;
}

static double now_cpu() {
  return ((double)clock()) / CLOCKS_PER_SEC;
}

static simMPI_Event* new_event(int myid) {
  int num;
  simMPI_Event *event, *old;

  num=simMPI[myid].EventNum;
  if (num==simMPI[myid].MAX_EVENT_NUM) {	/* expand */
    simMPI[myid].MAX_EVENT_NUM=2*num;
    old=simMPI[myid].event;
    simMPI[myid].event=(simMPI_Event*) anl_malloc (myid, simMPI[myid].MAX_EVENT_NUM*sizeof(simMPI_Event));
    memcpy(simMPI[myid].event, old, num*sizeof(simMPI_Event));
    anl_free(myid, old);
  }

  event = &(simMPI[myid].event[num]);
  simMPI[myid].EventNum = num+1;
  event->seq = (myid<<16) + num;
  event->size = 0;
  event->start_time = 0;
  return event;
}

/* bookkeeping in the beginning */
void sim_gstart(int myid) {
  simMPI[myid].clock = now_cpu();
}

static void sim_compute_end(int myid) {
  simMPI_Event *event=new_event(myid);
  event->type = Event_Compute;
  event->exec_time = (now_cpu() - simMPI[myid].clock)*COMPUTE;
  event->parents = 0;
  event->child_num = 0;
  event->children = NULL;
}

/* bookkeeping when an event starts */
/* type, parents, child_num and children will be decided later */
int sim_comm_start(int myid, int type, MPI_Comm comm, int root, int size) {
  int num;
  simMPI_Event *event;
  int nproc;

  sim_compute_end(myid);

  num = simMPI[myid].EventNum;
  event=new_event(myid);
  event->type = type;
  event->exec_time = 0;
  event->size = size;

  switch (type) {
  case Event_Send:
    event->parents=0;
    event->child_num=1;
    event->children=(int*)anl_malloc(myid, sizeof(int));
    break;
  case Event_Recv:
    event->parents=1;
    event->child_num=0;
    event->children=NULL;
    break;
  case Event_Bcast:
    MPI_Comm_size (comm, &nproc);
    if (root==myid) {	/* I am the root */
      event->parents=0;
      event->child_num=0;
      event->children=(int*)anl_malloc(myid, (nproc-1)*sizeof(int));
    }
    else {
      event->parents=1;
      event->child_num=0;
      event->children=NULL;
    }
    break;
  case Event_Barrier:
    MPI_Comm_size (comm, &nproc);
    event->parents=nproc-1;
    event->child_num=nproc-1;
    event->children=(int*)anl_malloc(myid, (nproc-1)*sizeof(int));
    break;
  }

  return num;
}

/* bookkeeping when an event ends */
void sim_comm_end(int myid) {
  simMPI[myid].clock = now_cpu();
}

/* bookkeeping in the end */
void sim_gend(int myid) {
  sim_compute_end(myid);
}

static double comm_time(int size) {
  return ALPHA+BELTA*size;
}

static void print_results (int nproc) {
  char fname[]="sim_results";
  FILE *fp;
  int i, j;

  fp = fopen (fname, "w");
  for (i=0; i<nproc; i++) {
    fprintf (fp, "## event sequence for node %d\n", i);
    for (j=0; j<simMPI[i].EventNum; j++) {
      switch (simMPI[i].event[j].type) {
      case Event_Compute: fprintf (fp, "    event #%d - Computation\n", j); break;
      case Event_Send: fprintf (fp, "    event #%d - Send\n", j); break;
      case Event_Recv: fprintf (fp, "    event #%d - Recv\n", j); break;
      case Event_Bcast: fprintf (fp, "    event #%d - Broadcast\n", j); break;	
      case Event_Barrier: fprintf (fp, "    event #%d - Barrier\n", j); break;	
      }
      fprintf (fp, "        start_time = %4.3f\n", simMPI[i].event[j].start_time);
      fprintf (fp, "        exec_time = %4.3f\n", simMPI[i].event[j].exec_time);
      fprintf (fp, "        %d incoming links, %d outgoing links\n", simMPI[i].event[j].parents, simMPI[i].event[j].child_num);
    }
    fprintf (fp, "\n");
  }
  fclose (fp);   
}

/* run by one node */
void sim_analyze(int nproc) {
  int node=0, noAdvance=0;
  int *pos, i, peer, lseq;
  double *ltime;
  double tmp_time;

  pos = (int*) malloc (nproc*sizeof(int));
  ltime = (double*) malloc (nproc*sizeof(double));
  for (i=0; i<nproc; i++) {
    pos[i]=0;
    ltime[i]=0;
  }
  
  /* the first event in each host is always a computation event */
  while (noAdvance<nproc) {
    if (pos[node]==simMPI[node].EventNum || simMPI[node].event[pos[node]].parents>0) {
      noAdvance++;
      node=(node+1)%nproc;
      continue;
    }
    ltime[node]=simMPI[node].event[pos[node]].start_time + simMPI[node].event[pos[node]].exec_time;

    noAdvance=0;
    pos[node]++;
    if (simMPI[node].event[pos[node]].start_time > ltime[node]) 
      ltime[node] = simMPI[node].event[pos[node]].start_time;
    else
      simMPI[node].event[pos[node]].start_time = ltime[node];

    for (i=0; i<simMPI[node].event[pos[node]].child_num; i++) {
      lseq=simMPI[node].event[pos[node]].children[i];
      peer=lseq>>16;
      lseq &= 0xffff;
      simMPI[peer].event[lseq].parents--;
      tmp_time = ltime[node] + comm_time(simMPI[node].event[pos[node]].size);
      if (simMPI[peer].event[lseq].start_time < tmp_time)
	simMPI[peer].event[lseq].start_time = tmp_time;
    }
  }

  print_results(nproc);
}

#endif


