/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

/* 
 * mbuffer_test.c -- using standard malloc to implement buffer management
 */

# define _REENTRANT
# include <stdlib.h>
#include "struct.h"
#include "function.h"
#include "globals.h"

#define Alignment 8

/**
 * The size field of BufferHandle has been changed to reflect the size of 
 * the whole buffer instead of the user seen buffer (which does not include
 * the size of the handle). 
 */

/* buffer handle */
typedef struct BufferHandle {
  struct BufferHandle *pre;
  struct BufferHandle *next;
  Boolean free;
  int size;
} BufferHandle;

# define BUF_HEADER_SIZE ((sizeof(BUF_HEADER_SIZE)+Alignment-1)/Alignment*Alignment)

#define MinFragment BUF_HEADER_SIZE+64

static BufferHandle **SysBuffer, **UsrBuffer;
static mutex_t *BufferMutex;
static cond_t *BufferCond;

/* message buffer initialization by whole system */
void mb_global_init(int nproc) {
  SysBuffer = (BufferHandle**) myMalloc (sizeof(BufferHandle*)*nproc);
  UsrBuffer = (BufferHandle**) myMalloc (sizeof(BufferHandle*)*nproc);
  BufferMutex = (mutex_t*) myMalloc (sizeof(mutex_t)*nproc);
  BufferCond = (cond_t*) myMalloc (sizeof(cond_t)*nproc);
}

/* message buffer initialization by each thread */
void mb_init(int myid, int bsize) {
  SysBuffer[myid] = (BufferHandle*) myMalloc (bsize);
  SysBuffer[myid]->pre = NULL;
  SysBuffer[myid]->next = NULL;
  SysBuffer[myid]->free = True;
  SysBuffer[myid]->size = bsize;

  UsrBuffer[myid] = NULL;

  mutex_init(&(BufferMutex[myid]));
  cond_init(&(BufferCond[myid]));
}

/* attach user defined buffer, implement MPI_Buffer_attach */
int MPI_Buffer_attach(void *buf, int bsize) {
  ARG *Args;
  int myid;

  Args=(ARG*)pthread_getspecific(key_data);
  myid=Args->id;

  if (UsrBuffer[myid] != NULL) return MPI_ERR_BUFFER;

  mutex_lock(&(BufferMutex[myid]));
  UsrBuffer[myid] = (BufferHandle*) buf;
  UsrBuffer[myid]->pre = NULL;
  UsrBuffer[myid]->next = NULL;
  UsrBuffer[myid]->free = True;
  UsrBuffer[myid]->size = bsize;
  mutex_unlock(&(BufferMutex[myid]));

  return MPI_SUCCESS;
}

/* detach user defined buffer, implement MPI_Buffer_detach */
int MPI_Buffer_detach(void *buf, int *bsize) {
  ARG *Args;
  int myid;

  Args=(ARG*)pthread_getspecific(key_data);
  myid=Args->id;

  if (UsrBuffer[myid]==NULL) return MPI_ERR_BUFFER;

  *((BufferHandle**)buf) = UsrBuffer[myid];
  *bsize = UsrBuffer[myid]->size;

  mutex_lock(&(BufferMutex[myid]));
  while (UsrBuffer[myid]->free==False || UsrBuffer[myid]->next!=NULL)
    cond_wait(&(BufferCond[myid]), &(BufferMutex[myid]));
  UsrBuffer[myid] = NULL;
  mutex_unlock(&(BufferMutex[myid]));

  return MPI_SUCCESS;
}

/* buffer allocation */
void *mb_alloc(int myid, int isize) {

  return memalign(Alignment, isize);
}

/* free an allocated buffer */
void mb_free(int myid, void* item) {
	free(item);
}
