/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

# define _REENTRANT
# include <stdio.h>
# include <stdlib.h>
# include <unistd.h>
# include "twait.h"
# include "tqueue.h"
# include "struct.h"
# include "mpi.h"
# include "mqueue.h"
# include "comm.h"
# include "mbuffer.h"
# include "root.h"

#ifdef MPISIM
#include "sim.h"
#endif

# if (_LOCK_)
# include "lock.h"
# endif

# if (_DEBUG_)
int err_ret[4][4], err_ts[4][4], err_stat[4][4];
thrMPI_SHandle *err_sh[4][4];

static void printSHandle(thrMPI_SHandle *sh, int t)
{
  if (t)
    fprintf(stderr, "## <NEW> Send handle=0x%x:\n", sh);
  else
    fprintf(stderr, "## Send handle=0x%x:\n", sh);

	fprintf(stderr, "  tag=%d\n", sh->tag);
	fprintf(stderr, "  src=%d\n", sh->src);
	fprintf(stderr, "  dest=%d\n", sh->dest);
	fprintf(stderr, "  datatype=%d\n", sh->datatype);
	fprintf(stderr, "  count=%d\n", sh->count);
	fprintf(stderr, "  buflen=%d\n", sh->buflen);
	fprintf(stderr, "  my_ts=%d\n", sh->my_ts);
}

static void printRHandle(thrMPI_RHandle *sh, int t)
{
  if (t)
    fprintf(stderr, "## <NEW> Recv handle=0x%x:\n", sh);
  else
    fprintf(stderr, "## Recv handle=0x%x:\n", sh);

	fprintf(stderr, "  tag=%d\n", sh->tag);
	fprintf(stderr, "  src=%d\n", sh->src);
	fprintf(stderr, "  dest=%d\n", sh->dest);
	fprintf(stderr, "  datatype=%d\n", sh->datatype);
	fprintf(stderr, "  count=%d\n", sh->count);
	fprintf(stderr, "  buflen=%d\n", sh->buflen);
	fprintf(stderr, "  my_ts=%d\n", sh->my_ts);
}
# endif

# define ASSERT(s) { if (!(s)) fprintf(stderr, "assertion failed\n"); }

extern int thrMPI_Nproc;

thrMPI_Channel **thrMPI_GCH=NULL;

/* check if it is a receive handle */
static int req_isRecv(MPI_Request req)
{
	if (req->type&(MPI_RECV|MPI_PERSISTENT_RECV|MPI_IMMEDIATE_RECV))
		return 1;
	return 0;
}

/* check if it is a send handle */
static int req_isSend(MPI_Request req)
{
	if (req->type&(MPI_SEND|MPI_PERSISTENT_SEND|MPI_IMMEDIATE_SEND))
		return 1;
	return 0;
}

/* check if it is a standard handle */
static int req_isStd(MPI_Request req)
{
	if (req->type&(MPI_SEND|MPI_RECV))
		return 1;
	return 0;
}

/* check if it is a persistent handle */
static int req_isPerm(MPI_Request req)
{
	if (req->type&(MPI_PERSISTENT_SEND|MPI_PERSISTENT_RECV))
		return 1;
	return 0;
}

/* check if it is an immediate handle */
static int req_isImmd(MPI_Request req)
{
	if (req->type&(MPI_IMMEDIATE_SEND|MPI_IMMEDIATE_RECV))
		return 1;
	return 0;
}

static int sq_cmp(thrMPI_SHandle *sh, int context, int tag, int status
#if (!_LOCK_)
, void *pts
#endif
)
{
#if (_DEBUG_)
  int sid = sh->src, rid = sh->dest;
#endif

	if ((sh->stat&status) && ((sh->tag==tag) || (tag==MPI_ANY_TAG)) 
		&& (sh->context==context) 
		) {
#if (!_LOCK_)
		*(int *)pts=sh->my_ts;
#endif
		return 0;
	}

	if (status & thrMPI_FREE) { /* searching for free handles */
		if (sh->stat & status)
			return 0;
		return 1;
	}

# if (_DEBUG_)
	err_sh[sid][rid] = sh;
	err_ts[sid][rid] = sh->my_ts;
	err_ret[sid][rid] = 4;	/* unreachable */
	if (!(sh->stat&status)) {
	  err_stat[sid][rid] = sh->stat;
	  err_ret[sid][rid] = 1;
	}
	if ( (sh->tag!=tag) && (tag!=MPI_ANY_TAG) )
		err_ret[sid][rid] = (sh->tag>tag)?2:-2;
	if (sh->context!=context)
		err_ret[sid][rid] = 3;
	return err_ret[sid][rid];
# else
	return 4;
# endif
}

static int rq_cmp(thrMPI_RHandle *rh, int context, int tag, int status
#if (!_LOCK_)
, void *pts
#endif
)
{
	if ((rh->stat&status) && ((rh->tag==tag) || (rh->tag==MPI_ANY_TAG)) 
		&& (rh->context==context) 
		) {
#if (!_LOCK_)
		*(int *)pts=rh->my_ts;
#endif
		return 0;
	}

	if (status & thrMPI_FREE) { /* searching for free handles */
		if (rh->stat & status)
			return 0;
		return 1;
	}

# if (_DEBUG_)

	if (!(rh->stat&status))
		return 1;
	if ( (rh->tag!=tag) && (rh->tag!=MPI_ANY_TAG) )
		return (rh->tag>tag)?2:-2;
	if (rh->context!=context)
		return 3;
	return 4;	/* unreachable */
# else
	return 4;
# endif
}

/* two stage initialization */
/* in main thread */
int thrMPI_GCH_ginit(void)
{
	int i=0, j;

	thrMPI_GCH=(thrMPI_Channel **)anl_malloc(ROOT_ID, sizeof(thrMPI_Channel *)*thrMPI_Nproc);
	if (thrMPI_GCH==NULL)
		return -1;

	thrMPI_GCH[0]=(thrMPI_Channel *)anl_malloc(ROOT_ID, sizeof(thrMPI_Channel)*thrMPI_Nproc*thrMPI_Nproc);
	if (thrMPI_GCH[0]==NULL)
		return -1;

# if (_SR_CHECK_)
	bzero(thrMPI_GCH[0], sizeof(thrMPI_Channel)*thrMPI_Nproc*thrMPI_Nproc);
# endif

	for (i=1; i<thrMPI_Nproc; i++)
		thrMPI_GCH[i]=thrMPI_GCH[i-1]+thrMPI_Nproc;

	return 0;
}

/* by each thread */
int thrMPI_GCH_uinit(int myid)
{
	int i,j ;

	for (i=0; i<thrMPI_Nproc; i++) {
# if (_LOCK_)
	  lock_init( &(thrMPI_GCH[myid][i].lock_flag));
# endif
		thrMPI_GCH[myid][i].sq=thrMPI_sq_create(thrMPI_Nproc, myid);
		thrMPI_GCH[i][myid].rq=thrMPI_rq_create(thrMPI_Nproc, myid);

		if ( (thrMPI_GCH[myid][i].sq==NULL) || (thrMPI_GCH[i][myid].rq==NULL) ) {
			for (j=0; j<i; j++) {
				Q_free(thrMPI_GCH[myid][j].sq);
				Q_free(thrMPI_GCH[j][myid].rq);
			}
			if (thrMPI_GCH[myid][i].sq)
				Q_free(thrMPI_GCH[myid][i].sq);
			if (thrMPI_GCH[i][myid].rq)
				Q_free(thrMPI_GCH[i][myid].rq);

			return -1;
		}
	}

	return 0;
}

TQueue *thrMPI_sq_create(int chunk_size, int owner)
{
	return Q_create((int (*)(void *, int, int, int
#if (!_LOCK_)
, void *
#endif
))sq_cmp, sizeof(thrMPI_SHandle), chunk_size+1, owner);
}

TQueue *thrMPI_rq_create(int chunk_size, int owner)
{
	return Q_create((int (*)(void *, int, int, int
#if (!_LOCK_)
, void *
#endif 
))rq_cmp, sizeof(thrMPI_RHandle), chunk_size+1, owner);
}

/* the following four function seems useless */
thrMPI_RHandle *rh_new(TQueue *rq)
{
	return (thrMPI_RHandle *)Q_newElem(rq);
}

thrMPI_SHandle *sh_new(TQueue *sq)
{
	return (thrMPI_SHandle *)Q_newElem(sq);
}

void rh_delete(TQueue *rq, thrMPI_RHandle *rh)
{
	rh->stat=thrMPI_DONE;
	Q_freeElem(rq, (void *)rh);
}

void sh_delete(TQueue *sq, thrMPI_SHandle *sh)
{
	sh->stat=thrMPI_DONE;
	Q_freeElem(sq, (void *)sh);
}

void sh_enQ(TQueue *sq, thrMPI_SHandle *sh, unsigned int init_stat)
{
#if (!_LOCK_)
	sh->my_ts=Q_now(sq);
	sh->peer_ts=0;
#endif
	sh->stat=init_stat;

	enQ(sq, (void *)sh);
}

void rh_enQ(TQueue *rq, thrMPI_RHandle *rh, unsigned int init_stat)
{
#if (!_LOCK_)
	rh->my_ts=Q_now(rq);
	rh->peer_ts=0;
#endif
	rh->stat=init_stat;
	enQ(rq, (void *)rh);
}

#if (!_LOCK_)
thrMPI_RHandle *rh_safe(TQueue *rq, thrMPI_RHandle *rh)
{
	thrMPI_RHandle *newrh;

	if (Q_isSafe(rq))
		return rh;

	newrh=rh_new(rq);
	memcpy((char *)newrh, (char *)rh, sizeof(thrMPI_RHandle));

	rh_delete(rq, rh);
	return newrh;
}

thrMPI_SHandle *sh_safe(TQueue *sq, thrMPI_SHandle *sh)
{
	thrMPI_SHandle *newsh;

	if (Q_isSafe(sq))
		return sh;

	newsh=sh_new(sq);
	memcpy((char *)newsh, (char *)sh, sizeof(thrMPI_SHandle));

	sh_delete(sq, sh);
	return newsh;
}
#endif

#if (!_LOCK_)
/* by sender */
thrMPI_RHandle *rq_probe(TQueue *rq, int context, int tag, int *ts)
{
	return (thrMPI_RHandle *)Q_probe(rq, context, tag, thrMPI_PEND|thrMPI_CREATE, (void *)ts);
}

thrMPI_RHandle *rq_probe_last(TQueue *rq, int context, int tag, int *ts, thrMPI_RHandle *last)
{
	return (thrMPI_RHandle *)Q_probe_last(rq, context, tag, \
		thrMPI_PEND|thrMPI_CREATE, (void *)ts, (void *)last);
}
# else 
thrMPI_RHandle *rq_probe(TQueue *rq, int context, int tag)
{
	return (thrMPI_RHandle *)Q_probe(rq, context, tag, thrMPI_PEND|thrMPI_CREATE);
}
#endif

#if (!_LOCK_)
/* by receiver */
thrMPI_SHandle *sq_probe(TQueue *sq, int context, int tag, int *ts)
{
	return (thrMPI_SHandle *)Q_probe(sq, context, tag, thrMPI_PEND|thrMPI_CREATE, (void *)ts);
}

thrMPI_SHandle *sq_probe_last(TQueue *sq, int context, int tag, int *ts, thrMPI_SHandle *last)
{
	return (thrMPI_SHandle *)Q_probe_last(sq, context, tag, \
		thrMPI_PEND|thrMPI_CREATE, (void *)ts, (void *)last);
}
#else
thrMPI_SHandle *sq_probe(TQueue *sq, int context, int tag)
{
	return (thrMPI_SHandle *)Q_probe(sq, context, tag, thrMPI_PEND|thrMPI_CREATE);
}
#endif

/* Three possible places where a handle can be unlinked from a queue: Garbage 
 * collection, Request free, MPI_Wait for persistent handle.
 */

/**
 * SHandles are freed asynchronuously while RHandles are freed synchronuously,
 * (at the point of the return of wait).
 * There could be 5*2*2=20 send possibilities:
 * For send mode, with the standard send has two variations: in-place buffering
 * or system buffer. Either immediate or not. And either persistent or not.
 * Under shmem architecture, data movement is done by memcpy, so essentially 
 * bsend and ibsend do the same thing. ... Eventually there are 6 
 * different possibilities.
 * +------------+---------------+----------------+--------+
 * |         	|  STD	 SYNC   |  STD   BUFFERD | READY  |
 * |            | USRBUF Issend	| SYSBUF Ibsend	 | Irsend |
 * +------------+---------------+----------------+--------+
 * | non-perm   |               |       +        |        |
 * +------------+---------------+----------------+--------+
 * |  perm     	|               |       +        |        |
 * +------------+---------------+----------------+--------+
 * Among these possibilities, only the two marked "+" will be freed asynchronuously.
 * To be specific, since IxSend will eventually be matched by a wait, so they will 
 * be treated as persistent handle by the garbage collector. But their wait behavior
 * will be different.
 * SHandle type: 1. basic, auto create, auto recycle: Send, Bsend, Rsend, Ssend
 *		 2. immediate, req start/wait: Isend, Ibsend, Irsend, Issend
 *		 3. permanent, req create, (start/wait)*, free: 
 * 			Send_Init, Bsend_Init, Rsend_Init, Ssend_Init.
 * RHandle type: 1. block, 2 immediate, 3 persistent
 * Buffer: 1. no-buffering: STD USRBUF, SYNC, READY
 * 	   2. buffering: STD SYSBUF, BUFFERED, 
 */
/* by receiver, garbage collection, should be nowhere to apply */

# if (_GC_) && (!_LOCK_)

void rq_GC(TQueue *rq)
{
	thrMPI_RHandle *rh=(thrMPI_RHandle *)Q_first(rq);
	int nfreed=0; /* number of contiguous free handles */
	int nremain=0; /* remaining free handles, may be inaccurate */
	int len;
	int i;

	while (rh && (rh->stat&(thrMPI_FREE|thrMPI_ERR)) ) {
		nfreed++;
		Q_removeElem(rq, rh);
		rh->stat=thrMPI_DONE;
		if (req_isStd((MPI_Request)rh)) {
			/* neither a persistent handle or an immediate handle */
			rh_delete(rq, rh);
		}
		rh=Q_next(rq, rh); /* this is correct! because next field is not changed */
	}

	nremain=Q_decr(rq, nfreed);
	len=Q_cnt(rq);

	if ( (nremain<=GC_MIN) || (len/nremain<=INUSE_FREE_UP) )
		return;

	nfreed=0;

	/* find and find_last will return an unlinked handle */
	rh=Q_find(rq, 0, -2, thrMPI_FREE|thrMPI_ERR, NULL);
	/* rh will not be NULL! because the ucnt is only decreased the master of the queue */

	while (1) {
		if (!rh) {
			fprintf(stderr, "Internal error, inconsistent ucnt state\n");
			fflush(stderr);
		}

		rh->stat=thrMPI_DONE;

		if (req_isStd((MPI_Request)rh)) {
			/* neither a persistent handle or an immediate handle */
			rh_delete(rq, rh);
		}
		nfreed++;
		nremain--;
		len--;

		rh=Q_next(rq, rh);

		while (rh && (rh->stat&(thrMPI_FREE|thrMPI_ERR)) ) {
			Q_removeElem(rq, rh);
			rh->stat=thrMPI_DONE;
			if (req_isStd((MPI_Request)rh)) {
				/* neither a persistent handle or an immediate handle */
				rh_delete(rq, rh);
			}
			nfreed++;
			nremain--;
			len--;
			rh=Q_next(rq, rh);
		}

		if ( rh && (nremain>GC_MIN) && (len/nremain>INUSE_FREE_LOW) )
			rh=Q_find_last(rq, 0, -2, (thrMPI_FREE|thrMPI_ERR), NULL, rh);
		else
			break;
	}

	Q_decr(rq, nfreed);

	return;
}

/* by sender, garbage collection */
void sq_GC(TQueue *sq)
{
	thrMPI_SHandle *sh=(thrMPI_SHandle *)Q_first(sq);
	int nfreed=0; /* number of contiguous free handles */
	int nremain=0; /* remaining free handles, may be inaccurate */
	int len;
	int i;

	while (sh && (sh->stat&(thrMPI_FREE|thrMPI_ERR) ) ) {
# if (_DEBUG_)
		if (!req_isSend((MPI_Request)sh))
			fprintf(stderr, "sq_GC: rhandle in sq!\n");
# endif
		nfreed++;
		Q_removeElem(sq, sh);
		sh->stat=thrMPI_DONE;
		if (req_isStd((MPI_Request)sh)) {
			/* neither a persistent handle or an immediate handle */
			/* free the buffer */
			/* I let the owner of the handle to free the buffer,
			 * this will defer the buffer deallocation than if we let the 
 			 * receiver to deallocate the buffer. The advantage is that
			 * we can split buffer into to parts, one for pt2pt communication
			 * and the other for group communication. The pt2pt part can
			 * be managed without any lock and the group communication requires
			 * either lock or equivalent lock-free methods.
			 */
			if (sh->useSysBuf==True) {
				if (sh->buflen)
					mb_free(sh->src, sh->bufadd);
			}

			sh_delete(sq, sh);
		}
		sh=Q_next(sq, sh); /* this is correct! because next field is not changed */
	}

	nremain=Q_decr(sq, nfreed);
	len=Q_cnt(sq);

	if ( (nremain<=GC_MIN) || (len/nremain<=INUSE_FREE_UP) )
		return;

	nfreed=0;

	/* find and find_last will return an unlinked handle */
	sh=Q_find(sq, 0, -2, thrMPI_FREE|thrMPI_ERR, NULL);
	/* sh will not be NULL! */

	while (1) {
# if (_DEBUG_)
		if (!sh) {
			fprintf(stderr, "Internal error, inconsistent ucnt state\n");
			fflush(stderr);
		}

		if (!req_isSend((MPI_Request)sh))
			fprintf(stderr, "sq_GC: rhandle in sq!\n");
# endif
		sh->stat=thrMPI_DONE;

		if (req_isStd((MPI_Request)sh)) {
			/* neither a persistent handle nor an immediate handle */
			/* free the buffer */
			if (sh->useSysBuf==True) {
				if (sh->buflen)
					mb_free(sh->src, sh->bufadd);
			}

			sh_delete(sq, sh);
		}
		nfreed++;
		nremain--;
		len--;

		sh=Q_next(sq, sh);

		while (sh && (sh->stat&(thrMPI_FREE|thrMPI_ERR)) ) {
			Q_removeElem(sq, sh);
			sh->stat=thrMPI_DONE;
			if (req_isStd((MPI_Request)sh)) {
				/* neither a persistent handle nor an immediate handle */
				/* free the buffer */
				if (sh->useSysBuf==True) {
					if (sh->buflen)
						mb_free(sh->src, sh->bufadd);
				}

				sh_delete(sq, sh);
			}
			nfreed++;
			nremain--;
			len--;
			sh=Q_next(sq, sh);
		}

		if ( sh && (nremain>GC_MIN) && (len/nremain>INUSE_FREE_LOW) )
			sh=Q_find_last(sq, 0, -2, thrMPI_FREE|thrMPI_ERR, NULL, sh);
		else
			break;
	}

	Q_decr(sq, nfreed);

	return;
}

#endif

void user_GC(int myid)
{
# if (_GC_) && (!_LOCK_)
	int i;

	for (i=0;i<thrMPI_Nproc; i++) {
		rq_GC(thrMPI_GCH[i][myid].rq);
		sq_GC(thrMPI_GCH[myid][i].sq);
	}
#endif
	return;
}



/* by sender */
int thrMPI_SH_init
(
MPI_Comm comm, 
char *buf, 
int len, 
int from, /* must be myid */
int to, 
int tag, 
int count,
thrMPI_Mode mode, 
Boolean isPersistent, 
Boolean isImmediate,  /* isImmediate and isPersistent cannot be both True, */
/* if they do, isImmediate has no effect */
MPI_Request *psh
)
{
	TQueue *sq=thrMPI_GCH[from][to].sq;
	thrMPI_SHandle *sh;
	thrMPI_Comm *tcomm=comm_verify(comm, from);

	if (len<0) {
		len=0;
	}

#if (_LOCK_)
	lock( &(thrMPI_GCH[from][to].lock_flag));
#endif	
	sh=sh_new(sq);
#if (_LOCK_)
	unlock( &(thrMPI_GCH[from][to].lock_flag));
#endif
	if (sh==NULL) {
		*psh=MPI_REQUEST_NULL;
# if (_DEBUG_)
		fprintf(stderr, "sh_init: no more memory\n");
# endif
		return MPI_ERR_INTERN;
	}

	if (isPersistent==True) {
		sh->handle_type=MPI_PERSISTENT_SEND;
	}
	else if (isImmediate==True) {
		sh->handle_type=MPI_IMMEDIATE_SEND;
	}
	else {
		sh->handle_type=MPI_SEND;
	}

	sh->cookie=MPI_REQUEST_COOKIE;

	sh->stat=thrMPI_DONE;

	if (tcomm==NULL) {
		sh->errval=MPI_ERR_COMM;
		*psh=(MPI_Request)sh;
# if (_DEBUG_)
		fprintf(stderr, "sh_init: wrong communicator\n");
# endif
		return MPI_ERR_COMM;
	}

	sh->context=tcomm->pt2pt_context; 
	sh->tag=tag;
	sh->src=from;
	sh->dest=to;
	sh->comm=comm;

	/* allocate buffer */
	sh->bufadd=NULL;

	if (mode&(STANDARD|BUFFERED)) {
		if (len)
			sh->bufadd=mb_alloc(from, len);
	}

	if ( (!sh->bufadd) && (len) ) { /* have to use the user in-place buffer */
		if (mode==BUFFERED) { /* unable to allocate buffer */
			sh->errval=MPI_ERR_BUFFER;
			*psh=(MPI_Request)sh;
# if (_DEBUG_)
			fprintf(stderr, "sh_init: no more buffer\n");
# endif
			return MPI_ERR_BUFFER;
		}

		sh->user_buf=sh->bufadd=buf;
		sh->useSysBuf=False;
# if (_DEBUG_)
		fprintf(stderr, "id=%d, len=%d, useSysBuf=False\n", from, len);
		fflush(stderr);
# endif
	}
	else {
		sh->useSysBuf=True;
		sh->user_buf=buf;
	}

	sh->buflen=len;
	sh->count=count;
	sh->mode=mode;

	/* sysbuf can be user attached or internal buffering */
	sh->active=False;
	sh->errval=MPI_SUCCESS;

	*psh=(MPI_Request)sh;

	return MPI_SUCCESS;
}

/* by receiver */
int thrMPI_RH_init
(
MPI_Comm comm, 
char *buf, 
int len, 
int from, 
int to, /* must be myid */
int tag, 
MPI_Datatype datatype,
int count,
Boolean isPersistent, 
Boolean isImmediate,  /* isImmediate and isPersistent cannot be both True, */
/* if they do, isImmediate has no effect */
/* MPI_Status *status, */ /* store the result */
MPI_Request *prh
)
{
	TQueue *rq=thrMPI_GCH[from][to].rq;
	thrMPI_RHandle *rh;
	thrMPI_Comm *tcomm=comm_verify(comm, to);

	if (from==MPI_ANY_SOURCE) {
		fprintf(stderr, "MPI_ANY_SOURCE not supported currently.\n");
		fflush(stderr);
		*prh=MPI_REQUEST_NULL;
		return MPI_ERR_INTERN;
	}

	if (len<0) {
		len=0;
	}

#if (_LOCK_)
	lock( &(thrMPI_GCH[from][to].lock_flag));
#endif
	rh=rh_new(rq);
#if (_LOCK_)
	unlock( &(thrMPI_GCH[from][to].lock_flag));
#endif
	
	if (rh==NULL) {
		*prh=MPI_REQUEST_NULL;
# if (_DEBUG_)
		fprintf(stderr, "rh_init: no more memory\n");
# endif
		return MPI_ERR_INTERN;
	}

	if (isPersistent==True) {
		rh->handle_type=MPI_PERSISTENT_RECV;
		/* this is because recv can use wildcard and the wildcard can be overriden */
		/* also actually received data can be less than specified in the request */
		rh->perm_source=from;
		rh->perm_tag=tag;
		rh->perm_len=len;
	}
	else if (isImmediate==True) {
		rh->handle_type=MPI_IMMEDIATE_RECV;
	}
	else {
		rh->handle_type=MPI_RECV;
	}

	rh->cookie=MPI_REQUEST_COOKIE;

	rh->stat=thrMPI_DONE; /* free and not in queue */

	if (tcomm==NULL) {
		rh->errval=MPI_ERR_COMM;
		*prh=(MPI_Request)rh;
# if (_DEBUG_)
		fprintf(stderr, "rh_init: wrong communicator\n");
# endif
		return MPI_ERR_COMM;
	}

	rh->context=tcomm->pt2pt_context; 
	rh->tag=tag;
	rh->src=from;
	rh->dest=to;
	rh->datatype=datatype;
	rh->count=count;
	rh->comm=comm;
	rh->bufadd=buf;
	rh->buflen=len;
	/* rh->status=status; */

	rh->active=False;
	rh->errval=MPI_SUCCESS;

	*prh=(MPI_Request)rh;

	return MPI_SUCCESS;
}

#if (_LOCK_)

/* will be called both internally and externally, it does some checking and then call */
/* internal sh_delete or rh_delete */
int thrMPI_Request_free(MPI_Request *req) {
  register TQueue *sq, *rq;
  register thrMPI_SHandle *sh;
  register thrMPI_RHandle *rh;

  if (*req==MPI_REQUEST_NULL) return MPI_SUCCESS;

  sh=&((*req)->shandle);

  if (!((*req)->common.stat&(thrMPI_ERR|thrMPI_DONE))) {
    /* could be only one possiblity: buffered send */
    /* remove this reference, and let it be freed by garbage collection */
    if ( (!req_isSend(*req)) || ((*req)->shandle.useSysBuf!=True) )
      return MPI_ERR_REQUEST;
    else {
      sh->I_have_no_interest = 1;
      *req=MPI_REQUEST_NULL;
      return MPI_SUCCESS;
    }
  }

  rh=&((*req)->rhandle);

  switch ((*req)->type) {
  case MPI_PERSISTENT_SEND: /* external */
    if (sh->active==True)
      return MPI_ERR_REQUEST;
  case MPI_IMMEDIATE_SEND: /* internal */
# if (_DEBUG_)
    fprintf(stderr, "why free an isend handle?\n");
    fflush(stderr);
# endif

  case MPI_SEND: /* internal */
    sq=thrMPI_GCH[sh->src][sh->dest].sq;
    
    /*    ASSERT(sh->stat&thrMPI_DONE); */

    if (sh->useSysBuf==True) {
      if (sh->buflen)
	mb_free(sh->src, sh->bufadd);
    }
    lock( &(thrMPI_GCH[sh->src][sh->dest].lock_flag));
    sh_delete(sq, sh);
    unlock( &(thrMPI_GCH[sh->src][sh->dest].lock_flag));
    break;

  case MPI_PERSISTENT_RECV: /* external */
    if (rh->active==True)
      return MPI_ERR_REQUEST;
  case MPI_RECV: /* internal */
  case MPI_IMMEDIATE_RECV: /* internal */
    rq=thrMPI_GCH[rh->src][rh->dest].rq;

    /*    ASSERT(sh->stat&thrMPI_DONE); */
    lock( &(thrMPI_GCH[rh->src][rh->dest].lock_flag));
    rh_delete(rq, rh);
    unlock( &(thrMPI_GCH[rh->src][rh->dest].lock_flag));
    break;
  default : 
    return MPI_ERR_REQUEST;
  }

  (*req)=MPI_REQUEST_NULL;
  return MPI_SUCCESS;
}

#else

/* will be called both internally and externally, it does some checking and then call */
/* internal sh_delete or rh_delete */
int thrMPI_Request_free(MPI_Request *req)
{
	register TQueue *sq, *rq;
	register thrMPI_SHandle *sh;
	register thrMPI_RHandle *rh;

	if (*req==MPI_REQUEST_NULL) return MPI_SUCCESS;


	if (!((*req)->common.stat&(thrMPI_FREE|thrMPI_ERR|thrMPI_DONE))) {
		/* could be only one possiblity: buffered send */
		/* remove this reference, and let it be freed by garbage collection */
		if ( (!req_isSend(*req)) || ((*req)->shandle.useSysBuf!=True) )
			return MPI_ERR_REQUEST;
		else {
			*req=MPI_REQUEST_NULL;
			return MPI_SUCCESS;
		}
	}

	sh=&((*req)->shandle);
	rh=&((*req)->rhandle);

	switch ((*req)->type) {
	case MPI_PERSISTENT_SEND: /* external */
		if (sh->active==True)
			return MPI_ERR_REQUEST;
	case MPI_IMMEDIATE_SEND: /* internal */
# if (_DEBUG_)
		fprintf(stderr, "why free an isend handle?\n");
		fflush(stderr);
# endif

	case MPI_SEND: /* internal */
		sq=thrMPI_GCH[sh->src][sh->dest].sq;

		if (sh->stat&(thrMPI_FREE|thrMPI_ERR)) { /* still linked in queue */
			Q_decr(sq, 1);
			Q_removeElem(sq, (void *)sh);
			sh->stat=thrMPI_DONE;
		}

		if (sh->useSysBuf==True) {
			if (sh->buflen)
				mb_free(sh->src, sh->bufadd);
		}

		sh_delete(sq, sh);
		break;

	case MPI_PERSISTENT_RECV: /* external */
		if (rh->active==True)
			return MPI_ERR_REQUEST;
	case MPI_RECV: /* internal */
	case MPI_IMMEDIATE_RECV: /* internal */
		rq=thrMPI_GCH[rh->src][rh->dest].rq;

		if (rh->stat&(thrMPI_FREE|thrMPI_ERR)) { /* still linked in queue */
			Q_decr(rq, 1);
			Q_removeElem(rq, (void *)rh);
			rh->stat=thrMPI_DONE;
		}

		rh_delete(rq, rh);
		break;
	default : 
		return MPI_ERR_REQUEST;
	}

	(*req)=MPI_REQUEST_NULL;
	return MPI_SUCCESS;
}

#endif

/* simple matching job */
/* assuming it is safe to manipulate rh and sh, i.e. sh must have MATCH stat and 
 * rh must have MATCH state or it is not linked in the queue */
static void recv_match(thrMPI_RHandle *rh, thrMPI_SHandle *sh, TQueue *sq)
{
# if (_DEBUG_)
  thrMPI_SHandle *newsh;
  int timestamp;
	if (sh->datatype!=rh->datatype) {
		fprintf(stderr, "recv_match: datatype mismatching: err_ret=%d, err_stat=%d, err_ts=%d, err_sh=0x%x\n", err_ret[sh->src][sh->dest], err_stat[sh->src][sh->dest], err_ts[sh->src][sh->dest], err_sh[sh->src][sh->dest]);
		printSHandle(sh, 0);
		printRHandle(rh, 0);
		newsh = sq_probe(sq, rh->context, rh->tag, &timestamp);
		printSHandle(newsh, 1);
		while(1) sleep(1);
	}
# endif

#ifdef MPISIM
	simMPI[sh->src].event[sh->lseq].children[0] = simMPI[sh->dest].event[rh->lseq].seq;
#endif
	rh->buflen=MIN(sh->buflen, rh->buflen);
	if (rh->buflen)
		memcpy(rh->bufadd, sh->bufadd, rh->buflen); /* memory copy */


	if (rh->tag==MPI_ANY_TAG)
		rh->tag=sh->tag;
	if (rh->src==MPI_ANY_SOURCE)
		rh->src=sh->src;
	rh->errval=MPI_SUCCESS;

#if (_LOCK_)
	/*	ASSERT(sh->stat&thrMPI_PEND); */
	Q_removeElem(sq, (void*)sh);
	sh->stat=thrMPI_DONE;
	if (sh->I_have_no_interest) {
	  if (sh->useSysBuf==True) {
	    if (sh->buflen)
	      mb_free_other(sh->src, sh->bufadd);
	  }
    
	  sh_delete(sq, sh);
	}
#else

	sh->stat=thrMPI_FREE; /* CANNOT set the flag immediately after the copy */
	/* fSignalFlag(rh->src, &(sh->stat), thrMPI_FREE); */
	Q_incr(sq, 1);

#endif

	/***
	if (rh->status) {
		rh->status->count=(rh->buflen/thrMPI_Usize(rh->datatype));
		rh->status->MPI_SOURCE=rh->src;
		rh->status->MPI_TAG=rh->tag;
		rh->status->MPI_ERROR=rh->errval;
	}
	***/

	return;
}

/* assuming it is safe to manipulate rh and sh, i.e. rh must have MATCH stat and 
 * sh must not be linked in the queue */
static void send_match1(thrMPI_RHandle *rh, thrMPI_SHandle *sh, TQueue *rq)
{
# if (_DEBUG_)  
  thrMPI_RHandle *newrh;
  int timestamp;
	if (sh->datatype!=rh->datatype) {
		fprintf(stderr, "send_match1: datatype mismatching:\n");
		printSHandle(sh, 0);
		printRHandle(rh, 0);
		newrh = rq_probe(rq, sh->context, sh->tag, &timestamp);
		printRHandle(newrh, 1);
		while(1) sleep(1);
	}

# endif

#ifdef MPISIM
	simMPI[sh->src].event[sh->lseq].children[0] = simMPI[sh->dest].event[rh->lseq].seq;
#endif
	rh->buflen=MIN(sh->buflen, rh->buflen);

	if (rh->buflen)
		memcpy(rh->bufadd, sh->user_buf, rh->buflen); /* memory copy */
	if (rh->tag==MPI_ANY_TAG)
		rh->tag=sh->tag;
	if (rh->src==MPI_ANY_SOURCE)
		rh->src=sh->src;
#if (_LOCK_)
	/*	ASSERT(rh->stat&thrMPI_PEND); */
	Q_removeElem(rq, (void*)rh);
	rh->stat=thrMPI_DONE;
#else

	rh->stat=thrMPI_FREE; /* CANNOT set the flag immediately after the copy */
	Q_incr(rq, 1);
#endif
	
	sh->errval=MPI_SUCCESS;

	/***
	if (rh->status) {
		rh->status->count=(rh->buflen/thrMPI_Usize(rh->datatype));
		rh->status->MPI_SOURCE=rh->src;
		rh->status->MPI_TAG=rh->tag;
		rh->status->MPI_ERROR=rh->errval; 
	}
	***/

	return;
}

/* assuming it is safe to manipulate rh and sh, i.e. rh must have MATCH stat and 
 * sh must have MATCH state */
static void send_match2(thrMPI_RHandle *rh, thrMPI_SHandle *sh, TQueue *rq)
{
# if (_DEBUG_)
  thrMPI_RHandle *newrh;
  int timestamp;
	if (sh->datatype!=rh->datatype) {
		newrh = rq_probe(rq, sh->context, sh->tag, &timestamp);
		fprintf(stderr, "send_match2: datatype mismatching:\n");
		printSHandle(sh, 0);
		printRHandle(rh, 0);
		printRHandle(newrh, 1);
		while(1) sleep(1);
	}
# endif

#ifdef MPISIM
	simMPI[sh->src].event[sh->lseq].children[0] = simMPI[sh->dest].event[rh->lseq].seq;
#endif
	rh->buflen=MIN(sh->buflen, rh->buflen);
	if (rh->buflen)
		memcpy(rh->bufadd, sh->bufadd, rh->buflen); /* memory copy */
	if (rh->tag==MPI_ANY_TAG)
		rh->tag=sh->tag;
	if (rh->src==MPI_ANY_SOURCE)
		rh->src=sh->src;

	rh->stat=thrMPI_FREE; /* CANNOT set the flag immediately after the copy */
	Q_incr(rq, 1);

	sh->errval=MPI_SUCCESS;

	/***
	if (rh->status) {
		rh->status->count=(rh->buflen/thrMPI_Usize(rh->datatype));
		rh->status->MPI_SOURCE=rh->src;
		rh->status->MPI_TAG=rh->tag;
		rh->status->MPI_ERROR=rh->errval;
	}
	***/

	return;
}

# if (_LOCK_) 
/* lock version of thrMPI_Start */
int thrMPI_Start(MPI_Request *req) {
        register TQueue *sq, *rq;
        register thrMPI_Common *common;
        register thrMPI_RHandle *rh;
        register thrMPI_SHandle *sh;

	if (*req==MPI_REQUEST_NULL) return MPI_SUCCESS;

        /* must start from state thrMPI_DONE */
	/*	ASSERT((*req)->common.stat & thrMPI_DONE); */

        common=&((*req)->common);

        rq=thrMPI_GCH[common->src][common->dest].rq;
        sq=thrMPI_GCH[common->src][common->dest].sq;

	if (req_isRecv(*req)) {

	  rh=&((*req)->rhandle);

# if (_SR_CHECK_)
		rh->seq_no=thrMPI_GCH[common->src][common->dest].r_no++;
# endif
	  lock( &(thrMPI_GCH[common->src][common->dest].lock_flag));
	  sh=sq_probe(sq, common->context, common->tag);

	  if (sh) {
	    /* got matched SHandle */
# if (_SR_CHECK_)
		if (rh->seq_no!=sh->seq_no) {
			fprintf(stderr, "Possible message miss match. S_no=%d, R_no=%d\n", sh->seq_no, rh->seq_no);
			fflush(stderr);
		}
# endif
	    recv_match(rh, sh, sq);
	  }
	  else {
	    rh_enQ(rq, rh, thrMPI_PEND);
	  }

	  unlock( &(thrMPI_GCH[common->src][common->dest].lock_flag));

	  (*req)=(MPI_Request)rh;
	  return MPI_SUCCESS;
	}
	else if (req_isSend(*req)) {
	  sh=&((*req)->shandle);

# if (_SR_CHECK_)
		sh->seq_no=thrMPI_GCH[common->src][common->dest].s_no++;
# endif

	  lock( &(thrMPI_GCH[common->src][common->dest].lock_flag));
	  rh=rq_probe(rq, common->context, common->tag);

	  if (rh) {
	    /* got matched RHandle */
# if (_SR_CHECK_)
		if (rh->seq_no!=sh->seq_no) {
			fprintf(stderr, "Possible message miss match. S_no=%d, R_no=%d\n", sh->seq_no, rh->seq_no);
			fflush(stderr);
		}
# endif
	    send_match1(rh, sh, rq);
	  }
	  else {
	    /* special handling for ready send */
	    if (sh->mode==READY) 
	      /* abort request */
	      sh->errval=MPI_ERR_PENDING;
	    else {
	      /* init buffer for once */
	      if (sh->useSysBuf) {
		if (sh->bufadd)
		  memcpy(sh->bufadd, sh->user_buf, sh->buflen);
	      }
	      sh->I_have_no_interest = 0;
	      sh_enQ(sq, sh, thrMPI_PEND);
	    }
	  }

	  unlock( &(thrMPI_GCH[common->src][common->dest].lock_flag));

	  (*req)=(MPI_Request)sh;
	  return MPI_SUCCESS;
	}
	else
	  return MPI_ERR_REQUEST;
}
# else

/* still some flaw in there, but have no time to fix */
/* we do need to implement ref-count, but currently it will not cause problem because */
/* we don't have MPI_Cancel, count is read by the ower of the queue and write by the visiter */
/* Q interface may need change, we can no longer return handles by return value */
/* the requests will be in either DONE or PENDING state */

/* BE CAUTIOUS: the req will be reset at the time of returning of thrMPI_Start */
int thrMPI_Start(MPI_Request *req)
{
	register TQueue *sq, *rq;
	register thrMPI_Common *common;
	register thrMPI_RHandle *rh;
	register thrMPI_SHandle *sh;
	int peer_ts;

	if (*req==MPI_REQUEST_NULL) return MPI_SUCCESS;

	/* must start from state thrMPI_DONE */
	if ( (*req)->common.stat !=thrMPI_DONE) {
# if (_DEBUG_)
		fprintf(stderr, "wrong request\n");
		fflush(stderr);
# endif
		return MPI_ERR_REQUEST;
	}

	common=&((*req)->common);

	rq=thrMPI_GCH[common->src][common->dest].rq;
	sq=thrMPI_GCH[common->src][common->dest].sq;

	if (req_isRecv(*req)) {
		if (req_isPerm(*req)) { /* persistent receive handle */
			rh=&((*req)->persistent_rhandle);
			if (rh->active==True) {
				/* may be a redundent check */
				rh->errval=MPI_ERR_REQUEST;
				return MPI_ERR_REQUEST;
			}
			rh->active=True; /* will be triggered off in MPI_Wait */
			rh=rh_safe(rq, rh); /* make sure the handle is safe to reuse */
			/* Q_safe(rq, (void *)rh);  */

			/* will be set in MPI_Wait call:
			rh->buflen=rh->perm_len;
			rh->tag=rh->perm_tag;
			rh->src=rh->perm_source; 
			*/

		}
		else {
			rh=&((*req)->rhandle);
		}

# if (_SR_CHECK_)
		rh->seq_no=thrMPI_GCH[common->src][common->dest].r_no++;
# endif

RESTART1:	
		sh=sq_probe(sq, common->context, common->tag, &peer_ts);
		while (sh)  {	/* found matching handle */
			/* wait till handle mature */
			while ((sh->stat==thrMPI_CREATE) && (sh->my_ts==peer_ts) ) {
				waitFlagNE((int *)&(sh->stat), thrMPI_CREATE, common->src);
			}

			if ( (sh->stat!=thrMPI_PEND) || (sh->my_ts!=peer_ts) ) {
				/* sorry, req may be cancelled or restarted */
				/* probe_last is not at all safe, but could be changed with */
				/* safe version */
				/* sh=sq_probe_last(sq, common->context, common->tag, sh); */
				goto RESTART1;
			}
			else {
				/* sh->stat==thrMPI_PEND && sh->my_ts==peer_ts */
				/* and they will not change, hopefully :-) */
				/* actually it may change after cancel is implemented */
				/* at that moment, we have to guarantee that sh will */
				/* not be reused until receiver has release the handle reference */
				if (__compare_and_swap(&(sh->stat), thrMPI_PEND, thrMPI_MATCH))
					break;
				else {
					/* probe_last is not at all safe */
					/* sh=sq_probe_last(sq, common->context, common->tag, sh); */
					goto RESTART1;
				}
			}
		}

		if (sh) {
			/* got matched SHandle */
# if (_DEBUG_)
			if (sh->my_ts != peer_ts) {
				fprintf(stderr, "timestampe changed, we've done a wrong thing\n");
				fflush(stderr);
				sh->stat=thrMPI_PEND;
				goto RESTART1;
			}
# endif
# if (_SR_CHECK_)
		if (rh->seq_no!=sh->seq_no) {
			fprintf(stderr, "Possible message miss match. S_no=%d, R_no=%d\n", sh->seq_no, rh->seq_no);
			fflush(stderr);
		}
# endif
			recv_match(rh, sh, sq);
			(*req)=(MPI_Request)rh;
			return MPI_SUCCESS;
		}
		else {
			/* rh->stat=thrMPI_CREATE; */ /* will be enQed to rq */
			rh_enQ(rq, rh, thrMPI_CREATE);

			/* check again */
			sh=sq_probe(sq, common->context, common->tag, &peer_ts);
			while (sh)  { /* false while, just to help break */
				/* found matching handle */
				rh->peer_ts=peer_ts;
				rh->stat=thrMPI_MATCH; /* now trying to match sh */


				/* wait till handle mature */
				while ( (sh->stat==thrMPI_CREATE) && (sh->my_ts==peer_ts) ) {
					waitFlagNE((int *)&(sh->stat), thrMPI_CREATE, common->src);
				}

				if ( (sh->stat==thrMPI_MATCH) && (sh->peer_ts==rh->my_ts) && (sh->my_ts==peer_ts) ) {
					/* peer also see the same thing */
					/* sender wait for stat change, receiver proceed */
					break;
				}
				else if ( (sh->stat==thrMPI_PEND) && (sh->my_ts==peer_ts) ) {
					/* receiver comes later */
					if (__compare_and_swap(&(sh->stat), thrMPI_PEND, thrMPI_MATCH))
						break;
				}

# if (_DEBUG_)
				if (rh->stat!=thrMPI_MATCH) {
					fprintf(stderr, "impossible things happened: rh->stat!=MATCH\n");
					fflush(stderr);
				}
# endif

				/* stat could be CANCEL or FREE or whatever */
				/* really complicated, send call cancelled */
				/* simply restart everything */
# if (_DEBUG_)
				fprintf(stderr, "recv_start, why you will reach here?\n");
                                fprintf(stderr, "My state is %d, peer state is %d\n", rh->stat, sh->stat);
                                fprintf(stderr, "My timestamp is %d, peer timestamp was %d, but is %d now. And peer's peer ts is %d.\n", rh->my_ts, rh->peer_ts, sh->my_ts, sh->peer_ts);
				fflush(stderr);
# endif
				Q_removeElem(rq, (void *)rh);
				rh->stat=thrMPI_DONE;
				rh=rh_safe(rq, rh);
				/* Q_safe(rq, (void *)rh); */

				/* just think probe_last is not safe */
				/* sh=sq_probe_last(sq, common->context, common->tag, sh); */

				goto RESTART1;
			}

			if (sh) {
# if (_DEBUG_)
				if (sh->my_ts != peer_ts) {
					fprintf(stderr, "timestampe changed, we've done a wrong thing\n");
					fflush(stderr);
					Q_removeElem(rq, (void *)rh);
					rh->stat=thrMPI_DONE;
					rh=rh_safe(rq, rh);
					sh->stat=thrMPI_PEND;
					goto RESTART1;
				}
# endif
				/* finally still got a match */
# if (_SR_CHECK_)
		if (rh->seq_no!=sh->seq_no) {
			fprintf(stderr, "Possible message miss match. S_no=%d, R_no=%d\n", sh->seq_no, rh->seq_no);
			fflush(stderr);
		}
# endif
				recv_match(rh, sh, sq);
				Q_removeElem(rq, (void *)rh);
				rh->stat=thrMPI_DONE;
				(*req)=(MPI_Request)rh;
				return MPI_SUCCESS;
			}

			/* indeed no match, rh is in the CREATE state */
			rh->stat=thrMPI_PEND; /* waiting for messages */
			(*req)=(MPI_Request)rh;
			return MPI_SUCCESS;
		}
	}
	else if (req_isSend(*req)) {
		int initBuf=1;

		if (req_isPerm(*req)) { /* persistent send handle */
			sh=&((*req)->persistent_shandle);
			if (sh->active==True) {
				/* may be a redundent check */
				sh->errval=MPI_ERR_REQUEST;
				return MPI_ERR_REQUEST;
			}
			sh->active=True; /* will be triggered off in MPI_Wait */
			/* Q_safe(sq, (void *)sh); */
			sh=sh_safe(sq, sh); /* make sure the handle is safe to reuse */
		}
		else {
			sh=&((*req)->shandle);
		}

# if (_SR_CHECK_)
		sh->seq_no=thrMPI_GCH[common->src][common->dest].s_no++;
# endif

RESTART2:	
		rh=rq_probe(rq, common->context, common->tag, &peer_ts);
		while (rh)  {	/* found matching handle */
			/* wait till handle mature */
			while ( (rh->stat==thrMPI_CREATE) && (rh->my_ts==peer_ts) ) {
				waitFlagNE((int *)&(rh->stat), thrMPI_CREATE, common->dest);
			}

			if ( (rh->stat!=thrMPI_PEND) || (rh->my_ts!=peer_ts) ) {
				/* sorry, req may be cancelled */
				/* probe_last is not at all safe */
				/* rh=rq_probe_last(rq, common->context, common->tag, rh); */
				goto RESTART2;
			}
			else {
				if (__compare_and_swap(&(rh->stat), thrMPI_PEND, thrMPI_MATCH))
					break;
				else {
					/* probe_last is not at all safe */
					/* rh=rq_probe_last(rq, common->context, common->tag, rh); */
					goto RESTART2;
				}
			}
		}

		if (rh) { /* got matched RHandle */
# if (_DEBUG_)
			if (rh->my_ts != peer_ts) {
				fprintf(stderr, "timestampe changed, we've done a wrong thing\n");
				fflush(stderr);
				rh->stat=thrMPI_PEND;
				goto RESTART2;
			}
# endif
# if (_SR_CHECK_)
		if (rh->seq_no!=sh->seq_no) {
			fprintf(stderr, "Possible message miss match. S_no=%d, R_no=%d\n", sh->seq_no, rh->seq_no);
			fflush(stderr);
		}
# endif
			send_match1(rh, sh, rq);
			(*req)=(MPI_Request)sh;
			return MPI_SUCCESS;
		}
		else {
			/* special handling for ready send */
			if (sh->mode==READY) { /* abort request */
				sh->errval=MPI_ERR_PENDING;
				(*req)=(MPI_Request)sh;
				return MPI_ERR_PENDING;
			}

			/* init buffer for once */
			if (sh->useSysBuf && initBuf) {
				if (sh->bufadd)
					memcpy(sh->bufadd, sh->user_buf, sh->buflen);
				initBuf=0;
			}

			/* sh->stat=thrMPI_CREATE; */ /* will be enQed to sq */
			sh_enQ(sq, sh, thrMPI_CREATE);

			/* check again */
			rh=rq_probe(rq, common->context, common->tag, &peer_ts);
			while (rh)  { /* false while, just to help break */
				/* found matching handle */
				sh->peer_ts=peer_ts;
				sh->stat=thrMPI_MATCH; /* now trying to match rh */
				/* from this point, receiver can proceed to any stage */
				/* including to the next receive */

				/* wait till handle mature */
				while ( (rh->stat==thrMPI_CREATE) && (rh->my_ts==peer_ts)) {
					waitFlagNE((int *)&(rh->stat), thrMPI_CREATE, common->dest);
				}

				if ( (rh->stat==thrMPI_MATCH) && (rh->peer_ts==sh->my_ts) && (rh->my_ts==peer_ts)) {
					/* peer also see the same thing */
					/* sender wait for stat change, receiver proceed */
					do {
						waitFlagNE((int *)&(sh->stat), thrMPI_MATCH, common->dest);
					} while (sh->stat==thrMPI_MATCH);
					/* ASSERT(sh->stat==thrMPI_FREE) */
					/* remove the request from the queue */
					Q_removeElem(sq, (void *)sh);
					Q_decr(sq, 1);
					sh->stat=thrMPI_DONE;
					sh->errval=MPI_SUCCESS;
					(*req)=(MPI_Request)sh;
					return MPI_SUCCESS;
				}
				else if ( (rh->stat==thrMPI_PEND) && (rh->my_ts==peer_ts) ) {
					/* sender comes later */
					if (__compare_and_swap(&(rh->stat), thrMPI_PEND, thrMPI_MATCH))
						break;
				}

				if (sh->stat!=thrMPI_MATCH) { /* meaning that */
					/* receiver does things really fast */
					/* it sets the MATCH flag and do all the copying */
					/* and set sh to be FREE, finally set rh to be DONE */
					/* ASSERT(sh->stat==thrMPI_FREE) */
					/* remove the request from the queue */
# if (_DEBUG_)
					fprintf(stderr, "send_start, peer moves really fast\n");
					fprintf(stderr, "My state is %d, peer state is %d\n", sh->stat, rh->stat);
					fprintf(stderr, "My timestamp is %d, peer timestamp was %d, but is %d now. And peer's peer ts is %d.\n", sh->my_ts, sh->peer_ts, rh->my_ts, rh->peer_ts);
					fflush(stderr);
# endif
					Q_removeElem(sq, (void *)sh);
					Q_decr(sq, 1);
					sh->stat=thrMPI_DONE;
					sh->errval=MPI_SUCCESS;
					(*req)=(MPI_Request)sh;
					return MPI_SUCCESS;
				}

				/* stat could be CANCEL or FREE or whatever */
				/* really complicated, send call cancelled */
				/* simply restart everything */
# if (_DEBUG_)
				fprintf(stderr, "send_start, why you will reach here?\n");
				fprintf(stderr, "My state is %d, peer state is %d\n", sh->stat, rh->stat);
				fprintf(stderr, "My timestamp is %d, peer timestamp was %d, but is %d now. And peer's peer ts is %d.\n", sh->my_ts, sh->peer_ts, rh->my_ts, rh->peer_ts);
				fflush(stderr);
# endif
				Q_removeElem(sq, (void *)sh);
				sh->stat=thrMPI_DONE;
				/* Q_safe(sq, (void *)sh); */
				sh=sh_safe(sq, sh);

				/* just think probe_last is not safe */
				/* sh=sq_probe_last(sq, common->context, common->tag, sh); */

				goto RESTART2;
			}

			if (rh) {
# if (_DEBUG_)
				if (rh->my_ts != peer_ts) {
					fprintf(stderr, "timestampe changed, we've done a wrong thing\n");
					fflush(stderr);
					Q_removeElem(sq, (void *)sh);
					sh->stat=thrMPI_DONE;
					/* Q_safe(sq, (void *)sh); */
					sh=sh_safe(sq, sh);
					rh->stat=thrMPI_PEND;
					goto RESTART2;
				}
				if (sh->stat!=thrMPI_MATCH) 
					fprintf(stderr, "send_start: trying to match a rhandle when my stat isnot match!\n");
# endif
				/* finally still got a match */
				send_match2(rh, sh, rq);
				Q_removeElem(sq, (void *)sh);
				sh->stat=thrMPI_DONE;
				(*req)=(MPI_Request)sh;
				return MPI_SUCCESS;
			}

			/* indeed no match, sh is in the CREATE state */
			sh->stat=thrMPI_PEND; /* waiting for messages */
			(*req)=(MPI_Request)sh;
			return MPI_SUCCESS;
		}
	}
	else
		return MPI_ERR_REQUEST;
}
#endif


int thrMPI_Waitall( int count, MPI_Request *array_of_requests, MPI_Status *array_of_statuses) {
	register int i;
	for (i=0; i<count; i++)
		thrMPI_Wait( &array_of_requests[i], &array_of_statuses[i]);
	return MPI_SUCCESS;
}


#if (_LOCK_)

int thrMPI_Wait(MPI_Request *req, MPI_Status *status) {
  register thrMPI_SHandle *sh;
  register thrMPI_RHandle *rh;
  register thrMPI_Common *common;
  register TQueue  *sq;
  register TQueue  *rq;
  int ret;

  /* MPI_REQUEST_NULL is a legal request */
  if (*req==MPI_REQUEST_NULL) return MPI_SUCCESS;

  common=&((*req)->common);

  if (req_isRecv(*req)) {
    rh=&((*req)->rhandle);

    while (!(rh->stat&(thrMPI_DONE|thrMPI_FREE|thrMPI_ERR))) {
      waitFlagIN((int *)&(rh->stat), thrMPI_DONE|thrMPI_FREE|thrMPI_ERR, common->src);
    }

    /* now everything is OK */
    if (status) {
      status->buflen=rh->buflen;
      status->MPI_SOURCE=rh->src;
      status->MPI_TAG=rh->tag;
      status->MPI_ERROR=rh->errval;
    }

    ret=rh->errval;

    thrMPI_Request_free(req);

    return ret;
  }
  else if (req_isSend(*req)) {
    sq=thrMPI_GCH[common->src][common->dest].sq;

    sh=&((*req)->shandle);
    
    if (sh->stat&(thrMPI_DONE|thrMPI_ERR)) {
      /* no need to wait */
    }
    else if (sh->useSysBuf==True) {
      /* also don't need to wait */
    }
    else {
      do {
	waitFlagIN((int *)&(sh->stat), thrMPI_DONE|thrMPI_ERR, common->dest);
      } while (!(sh->stat&(thrMPI_DONE|thrMPI_ERR)));
    }

    /* now sh can be freed by garbage collector */
    if (sh->handle_type==MPI_IMMEDIATE_SEND) {
      sh->handle_type=MPI_SEND;
    }

    /* now everything is OK */
    if (status) {
      status->buflen=sh->buflen;
      status->MPI_SOURCE=sh->src; /* not necessary */
      status->MPI_TAG=sh->tag;
      status->MPI_ERROR=sh->errval;
    }

    ret=sh->errval;

    thrMPI_Request_free(req);

    return ret;
  }
  else
    return MPI_ERR_REQUEST;
}

#else

int thrMPI_Wait(MPI_Request *req, MPI_Status *status)
{
	register thrMPI_SHandle *sh;
	register thrMPI_RHandle *rh;
	register thrMPI_Common *common;
	register TQueue  *sq;
	register TQueue  *rq;
	int ret;

	/* MPI_REQUEST_NULL is a legal request */
	if (*req==MPI_REQUEST_NULL) return MPI_SUCCESS;

	common=&((*req)->common);

	if (req_isRecv(*req)) {
		/**
		if (isPerm(*req)) {
			rh=&((*req)->persistent_rhandle);
		}
		else
		**/
		rh=&((*req)->rhandle);

		while (!(rh->stat&(thrMPI_DONE|thrMPI_FREE|thrMPI_ERR))) {
		  waitFlagIN((int *)&(rh->stat), thrMPI_DONE|thrMPI_FREE|thrMPI_ERR, common->src);
		}

		/* now everything is OK */
		if (status) {
			status->buflen=rh->buflen;
			status->MPI_SOURCE=rh->src;
			status->MPI_TAG=rh->tag;
			status->MPI_ERROR=rh->errval;
		}

		ret=rh->errval;

		if (!req_isPerm(*req)) {
			thrMPI_Request_free(req);
		}
		else { /* reset the handle status */
			if (rh->stat&(thrMPI_FREE|thrMPI_ERR)) {
				rq=thrMPI_GCH[common->src][common->dest].rq;
				Q_removeElem(rq, (void *)rh);
				Q_decr(rq, 1);
				rh->stat=thrMPI_DONE;
			}

			rh->buflen=rh->perm_len;
			rh->tag=rh->perm_tag;
			rh->src=rh->perm_source;
			rh->active=False;
			rh->errval=MPI_SUCCESS;
		}
		return ret;
	}
	else if (req_isSend(*req)) {
		sq=thrMPI_GCH[common->src][common->dest].sq;

		/**
		if (isPerm(*req)) {
			sh=&((*req)->persistent_shandle);
		}
		else
		**/
		sh=&((*req)->shandle);

		if (sh->stat&(thrMPI_DONE|thrMPI_FREE|thrMPI_ERR)) {
			/* no need to wait */
		}
		else if (sh->useSysBuf==True) {
			/* also don't need to wait */
		}
		else {
			do {
				waitFlagIN((int *)&(sh->stat), \
					thrMPI_DONE|thrMPI_FREE|thrMPI_ERR, common->dest);
			} while (!(sh->stat&(thrMPI_DONE|thrMPI_FREE|thrMPI_ERR)));
		}

		/* now sh can be freed by garbage collector */
		if (sh->handle_type==MPI_IMMEDIATE_SEND) {
			sh->handle_type=MPI_SEND;
		}

		/* now everything is OK */
		if (status) {
			status->buflen=sh->buflen;
			status->MPI_SOURCE=sh->src; /* not necessary */
			status->MPI_TAG=sh->tag;
			status->MPI_ERROR=sh->errval;
		}

		ret=sh->errval;

		if (!req_isPerm(*req)) {
			thrMPI_Request_free(req);
		}
		else { /* reset the handle status */
			if (sh->stat&(thrMPI_FREE|thrMPI_ERR)) {
				Q_removeElem(sq, (void *)sh);
				Q_decr(sq, 1);
				sh->stat=thrMPI_DONE;
			}
			sh->active=False;
			if (sh->stat==thrMPI_DONE) { /* reuse the same handle */
				sh->errval=MPI_SUCCESS;
			}
			else { /* persistent buffer handling */
				/* replicate a new handle */
				thrMPI_SHandle *newsh;
				newsh=sh_new(sq);
				if (!newsh) {
					/* can insert error recovery here */
					if (status)
						status->MPI_ERROR=sh->errval=MPI_ERR_INTERN;
					ret=MPI_ERR_INTERN;
				}
				else {
					memcpy((char *)newsh, (char *)sh, \
						sizeof(thrMPI_SHandle));
					if (sh->buflen)
						newsh->bufadd=mb_alloc(sh->src, sh->buflen);
					else 
						newsh->bufadd=NULL;

					if ((!newsh->bufadd) && (sh->buflen) ) {
						if (sh->mode==STANDARD) {
							newsh->bufadd=newsh->user_buf;
							newsh->useSysBuf=False;

							newsh->stat=thrMPI_DONE;
							newsh->errval=MPI_SUCCESS;
							sh->handle_type=MPI_SEND;
							*req=(MPI_Request)newsh;
						}
						else {
							/* can insert error recovery here */
							sh_delete(sq, sh);
							if (status)
								status->MPI_ERROR=sh->errval=\
									MPI_ERR_INTERN;
							ret=MPI_ERR_INTERN;
						}
					}
					else {
						newsh->stat=thrMPI_DONE;
						newsh->errval=MPI_SUCCESS;
						sh->handle_type=MPI_SEND;
						*req=(MPI_Request)newsh;
					}
				}
			}
		}
#if (_GC_)
		sq_GC(sq);
#endif

		return ret;
	}
	else
		return MPI_ERR_REQUEST;
}

#endif
