/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

/**
 * This file is a simplfied version of the hash table implementation 
 * by Allen I. Holub. 
 * The bigest difference is the "cmp" function used in findsym. It
 * does not take two symbol structure, instead, it takes two keys.
 * findsym also uses a key to find desired symbol instead of a symbol.
 * Thus the maketab requires one more function parameter which
 * extracts the key from the symbol structure.
 */
# define _REENTRANT
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "tqueue.h"
#include "twait.h"
#include "anl_shmem.h"

# define ASSERT(s) { if (!(s)) fprintf(stderr, "assertion failed\n"); }

# if (_DEBUG_)
int maxQlen=0;
# endif

#if (!_LOCK_)	

void Q_freeAll(TQueue *p)
{
	register BUCKET *freeend=p->freed;
	register BUCKET *e;

	if (freeend) {
		while (freeend->prev)
			freeend=freeend->prev;

		freeend->prev=p->allocate;
		p->allocate=p->freed;
		p->freed=NULL;
	}

	return;
}

int Q_isSafe(TQueue *p)
{
	if (p->nTraverse>0) return 0;
# if (_DEBUG_)
	if (p->nTraverse<0)
		fprintf(stderr, "nTraverse<0? (%d) \n", p->nTraverse);
# endif

	Q_freeAll(p);

	return 1;
}

/* the only function that may block */
int Q_safe(TQueue *p, void *elem)
{
	
	int i, q;

# if (_DEBUG_)
	BUCKET *e=(BUCKET *)elem-1;
	
	if (e->prev!=(BUCKET *)1) {
		fprintf(stderr, "assuming a linked handle as freed\n");
	}
# endif
	i=64*1000;
	while ((i--) && ((q=p->nTraverse)>0));
	
	if (!q) return 0;

	i=100;
	while ((i--) && ((q=p->nTraverse)>0)) sched_yield();

	if (!q) return 0;

# if (_DEBUG_)
	fprintf(stderr, "may block in Q_safe\n");
# endif

	while (p->nTraverse>0) {
# if (_DEBUG_)
		fprintf(stderr, "nTraverse=%d\n", p->nTraverse);
# endif
		usleep(1000);
	}

# if (_DEBUG_)
	fprintf(stderr, "Q_safe succeded\n");
# endif

	Q_freeAll(p);

	return 0;
}	

#endif /* _LOCK_ */

/* master operation */
void *Q_newElem(TQueue *p)
{

 	BUCKET *elem;
	register BUCKET *b, *b1;
	register int size1;

	size1=p->item_size;

	if (!p->allocate) {
		int i;

 		if (!(b=elem=(BUCKET *)anl_malloc(p->owner, size1*p->chunk_group))) {
			return NULL;
 		}

		b1=(BUCKET *)((char *)b+size1);
		for (i=1; i<p->chunk_group-1; i++) {
			b=b1;
			b1=(BUCKET *)((char *)b+size1);
			b->prev=b1;
		}
			
		b1->prev=NULL;

		p->allocate=(BUCKET *)((char *)elem+size1);
	}
	else {
		elem=p->allocate;
		
		p->allocate=p->allocate->prev;
	}

# if (_DEBUG_)
	if (elem == &(p->qctrl)) fprintf(stderr, "Die here");
	elem->prev=(BUCKET *)1;
# endif

	elem->next=NULL;

 	return (void *)(elem+1); /* return pointer to user space */
}

/* master operation */
void Q_freeElem(TQueue *p, void *elem)
{
	BUCKET *e;

	if (!elem) return;

	e=(BUCKET *)elem-1;

#if (_LOCK_)

	e->prev = p->allocate;
	p->allocate = e;

#else

		e->prev=p->freed;
		p->freed=e;

	if (p->nTraverse<=0) { /* no traversal proc stations at elem */
# if (_DEBUG_)
		if (p->nTraverse<0)
			fprintf(stderr, "nTraverse <0 (%d)\n", p->nTraverse);
# endif
		Q_freeAll(p);
	}

#endif /* _LOCK_ */
}

TQueue * Q_create( int (* cmp_function)(void *elem, int context, int tag, int arg
#if (!_LOCK_)
, void *arg2
#endif
), int item_size, int group, int owner)
{
	
 	TQueue *p;
	
	if ( (item_size <= 0) || (group <= 1) ) 
		return NULL;

 	if (p=(TQueue *)anl_malloc(owner, sizeof(TQueue))) {
		Q_init(p, cmp_function, item_size, group, owner); 
 	}
 	else {
		return NULL;
 	}
 	return p;
}

/* p must be allocatd by Q_create */
void Q_free(TQueue *p)
{
	anl_free(p->owner, p); /* all handle memory are lost! */
}

BUCKET *Q_delElem(TQueue *p, BUCKET *elem)
{
# if (_DEBUG_)
	if (elem->prev==(BUCKET *)1) {
		fprintf(stderr, "attempting to remove a free handle\n");
		/* return NULL; */
	}
# endif

	elem->prev->next=elem->next;
	elem->next->prev=elem->prev;


# if (_DEBUG_)
	elem->prev=(BUCKET *)1;
# endif
	p->count--;

/**
	if (p->nTraverse>0) {
		elem->prev=p->unsafe;
		p->unsafe=elem;
	}
	else {
		elem->prev=(BUCKET *)1;

		Q_freeAll(p);
	}
**/

	return elem;
}

/* master operation */
void *Q_removeElem(TQueue *p, void *elem)
{
	Q_delElem(p, (BUCKET *)elem-1);

	return elem;
}

/* master operation */
void Q_init( TQueue *p, int (* cmp_function)(void *elem, int context, int tag, int arg
#if (!_LOCK_)
, void *arg2
#endif
), int item_size, int group, int owner)
{
	p->count = 0;
	p->ucount = 0;
	p->nts=0;
	p->cmp = cmp_function;
	p->qctrl.prev=p->qctrl.next=&(p->qctrl);
	p->freed=NULL;
	/* p->unsafe=NULL; */
	p->nTraverse=0;
	p->allocate=NULL;
	p->owner=owner;

	/* item_size must be a multiple of 8 */
 	/* BUCKET is already of size multiple of 8 */
	if (item_size % 0x07)
		p->item_size=(item_size+7)>>3<<3;
	else	
		p->item_size=item_size;

	p->item_size+=sizeof(BUCKET);
	p->chunk_group=group;

	Q_freeElem(p, Q_newElem(p)); /* first touch to generate initial handle cache */

	return;
}

#if (!_LOCK_)
	
/* master operation */
void Q_reset( TQueue *p)
{
	
	p->nTraverse=0;

	while ( &(p->qctrl) != p->qctrl.next )
		Q_freeElem(p, Q_removeElem(p, p->qctrl.next));
	Q_freeAll(p); /* just to make sure */

	p->nTraverse=0;
	p->ucount=0;
	p->nts=0;
}

#endif

/* master operation */
void enQ(TQueue *p, void *elem)
{
	BUCKET *e=(BUCKET *)elem-1;
	
# if (_DEBUG_)
	if (e->prev!=(BUCKET *)1) {
		fprintf(stderr, "attempting to enQ a non-free handle\n");
		return;
	}
# endif

	e->next=&(p->qctrl);
	e->prev=p->qctrl.prev;
	p->qctrl.prev->next=e;
	p->qctrl.prev=e;
	p->count++;

# if (_DEBUG_)
	if (p->count>maxQlen)
		maxQlen=p->count;
# endif
	
	return;
}
	
/* master operation */
void * deQ(TQueue *p)
{

	if ( p->qctrl.next==&(p->qctrl) )
		return NULL;
	return (void *)(Q_delElem(p, p->qctrl.next)+1);
}

void Q_status(TQueue *p)
{
	int nfreed=0, nunsafe=0, nalloc=0;
	BUCKET *e;

	e=p->freed;
	while (e) {
		nfreed++;
		e=e->prev;
	}

	e=p->allocate;
	while (e) {
		nalloc++;
		e=e->prev;
	}

	printf("Queue: \n\tQ->count = %d;\n\tQ->ucount = %d;\n\tQ->freed=%d;\n\tQ->allocate=%d;\n", p->count, p->ucount, nfreed, nalloc);

	return;
}

#if (!_LOCK_)
void printQ(TQueue *p)
{
	BUCKET *e=p->qctrl.next;
	
	if (e==&(p->qctrl)) {
		printf("empty queue\n");
		ASSERT(Q_isEmpty(p));
		return;
	}
	printf("Queue: Head");
	while (e!=&(p->qctrl)) {
		printf("-->%d", *(int *)(e+1));
		e=e->next;
	}
	printf("<--Tail\n");
}
# endif

void *Q_probe(TQueue *p, int context, int tag, int arg
#if (!_LOCK_)
, void *arg2
#endif
)
{
	register BUCKET *e;
	register BUCKET *f;
	int done=0;
	
	if (!p->cmp) return NULL;
	
#if (!_LOCK_)
	p->nTraverse++; 
#endif
	f=&(p->qctrl);
	e=p->qctrl.next;
#if (_DEBUG_)
	if (e != p->qctrl.next) fprintf(stderr, "sth scary happened!\n");
#endif
	while (e && (e!=f) ) {
		if ( (*p->cmp)((void *)(e+1), context, tag, arg
#if (!_LOCK_)
, arg2
#endif
)==0) {

			done=1;
			break;
		}
		e=e->next;
	}

#if (!_LOCK_)
	p->nTraverse--;
#endif
	if (!done) return NULL;
	else return (void *)(e+1); 
}

void Q_probe_last_init(TQueue *p)
{
	/* __add_and_fetch(&(p->nTraverse), 1); */
	p->nTraverse++;
}

void Q_probe_last_finish(TQueue *p)
{
	/* __sub_and_fetch(&(p->nTraverse), 1); */
	p->nTraverse--; 
}

#if (!_LOCK_)

/* User must ensure elem is really in the queue */
void *Q_probe_last(TQueue *p, int context, int tag, int arg, void *arg2, void *elem)
{
	BUCKET *e=((BUCKET *)elem-1)->next;
	int done=0;

	if (!p->cmp) return NULL;

	while (e && (e!=&(p->qctrl))) {
		if ( (*p->cmp)((void *)(e+1), context, tag, arg, arg2)==0) {
			done=1;
			break;
		}
		e=e->next;
	}
	if (!done) return NULL;
	else return (void *)(e+1);
}

/* master operation */
void *Q_find(TQueue *p, int context, int tag, int arg, void *arg2)
{
	register BUCKET *f=&(p->qctrl), *e=p->qctrl.next;
	
	if (!p->cmp) return NULL;

	while (e && (e!=f) ) {
		if ( (*p->cmp)((void *)(e+1), context, tag, arg, arg2)==0) {
			Q_delElem(p, e);
			return (void *)(e+1);
		}
		e=e->next;
	}

	return NULL;
}

/* master operation, elem could even be the one returned by last Q_find/Q_find_last */
void *Q_find_last(TQueue *p, int context, int tag, int arg, void *arg2, void *elem)
{
	register BUCKET *f=&(p->qctrl), *e=((BUCKET *)elem-1)->next;

	if (!p->cmp) return NULL;

	while (e && (e!=f) ) {
		if ( (*p->cmp)((void *)(e+1), context, tag, arg, arg2)==0) {
			Q_delElem(p, e);
			return (void *)(e+1);
		}
		e=e->next;
	}

	return NULL;
}

int Q_isEmpty(TQueue *q)
{
	if (q->count) return 0;
	return 1;

	/**
	if (q->qctrl.next == &(q->qctrl) ) return 1;
	else return 0;
	**/
}

void *Q_first(TQueue *p)
{
	register BUCKET *e=&(p->qctrl);
	register BUCKET *f=e->next;

	return (e==f)? NULL: (void *)(f+1);
}

void *Q_next(TQueue *p, void *item)
{
	register BUCKET *e=&(p->qctrl);
	register BUCKET *f=((BUCKET *)item-1)->next;

	return (e==f)? NULL : (void *)(f+1);
}

void *Q_last(TQueue *p)
{
	register BUCKET *e=&(p->qctrl);
	register BUCKET *f=e->prev;

	return (e==f)? NULL: (void *)(f+1);
}

int Q_incr(TQueue *p, int n)
{
	if (n)
		return __add_and_fetch(&(p->ucount), n);
	else
		return p->ucount;
}

int Q_decr(TQueue *p, int n)
{
	if (n)
		return __sub_and_fetch(&(p->ucount), n);
	else
		return p->ucount;
}

int Q_ucnt(TQueue *p)
{
	return p->ucount;
}

void Q_resetUcnt(TQueue *p)
{
	p->ucount=0;
	return;
}

int Q_cnt(TQueue *p)
{
	return p->count;
}

int Q_now(TQueue *p)
{
	return p->nts++;
}

#endif /* _LOCK_ */
