/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <strings.h>

#include "mytp.h"

#define SHM_KEY 111
#define SHM_SIZE 4096
#define MAX_APP 4
#define NUM_PROCS 4	/* number of processors in this machine */

static int shmid;
static void *shmvaddr;

void shmem_init (size_t shmsize) {
  if ((shmid = shmget (SHM_KEY, shmsize, IPC_CREAT|SHM_R|SHM_W|044)) == -1) {
    perror ("shmget");
    exit (1);
  }
	
  if ((shmvaddr = shmat (shmid, (void *)0, 0)) == (void *)(-1)) {
    perror ("shmat");
    exit (1);
  }

#ifdef DEBUG
  printf ("shmem_init: shmid=0x%x shmvaddr=0x%x\n", shmid, shmvaddr); 
  fflush (stdout);
#endif
}

void shmem_end () {
  if (shmdt (shmvaddr) == -1) {
    perror ("shmdt");
    exit (1);
  }
  /*
  shmctl (shmid, IPC_RMID);
  */
}

/* data structure for shmem area */

typedef struct {
  int total;
  int running;
  int recommended;
} app_t;

typedef struct {
  int daemonlock;
  int sys_total;
  int sys_running;
  app_t app[MAX_APP+1];
} daemon_t;

/* redistribute processors among applications */
static void redistribution () {
  int allocated=0, i, index;
  app_t *app_ptr;

  for (i=0; i<MAX_APP; i++) ((daemon_t*)shmvaddr)->app[i].recommended=0;

  i=0;
  while (allocated < NUM_PROCS) {
    index = i%MAX_APP; i++;
    app_ptr = ((daemon_t*)shmvaddr)->app + index;
    if ((app_ptr->total != 0) && (app_ptr->recommended < app_ptr->total)) {
      app_ptr->recommended ++; allocated++;
    }
  }
}

#ifdef DEBUG

static void dump_shmem () {
  int i;
  app_t *app_ptr;

  printf ("  ## shmem content: daemonlock=%d, sys_total=%d, sys_running=%d\n", ((daemon_t*)shmvaddr)->daemonlock, ((daemon_t*)shmvaddr)->sys_total, ((daemon_t*)shmvaddr)->sys_running);
  for (i=0; i<MAX_APP; i++) {
    app_ptr = ((daemon_t*)shmvaddr)->app + i;
    if (app_ptr->total != 0) 
      printf ("    APP %d: total=%d, running=%d, recommended=%d\n", i, app_ptr->total, app_ptr->running, app_ptr->recommended);
  }
  fflush (stdout);
}

#endif

/* register an application - return its id */
int app_register (int num_threads, int first) {
  int i=0;

  shmem_init (SHM_SIZE);
  if (first) bzero (shmvaddr, sizeof(daemon_t));

#ifdef DEBUG
  printf ("begin register - %d proc\n", num_threads);
  dump_shmem ();
#endif

  mytp_lock (& (((daemon_t*)shmvaddr)->daemonlock));

  while (((daemon_t*)shmvaddr)->app[i].total != 0) i++;
  if (i==MAX_APP) perror ("too many concurrent application");

  ((daemon_t*)shmvaddr)->app[i].total = ((daemon_t*)shmvaddr)->app[i].running = ((daemon_t*)shmvaddr)->app[i].recommended = num_threads;
  ((daemon_t*)shmvaddr)->sys_total += num_threads;
  ((daemon_t*)shmvaddr)->sys_running += num_threads;

  if (((daemon_t*)shmvaddr)->sys_running > NUM_PROCS) 
    redistribution();

  mytp_unlock (& (((daemon_t*)shmvaddr)->daemonlock));

#ifdef DEBUG
  printf ("end register\n");
  dump_shmem ();
#endif

  return i;
}

/* remove an application - specified by aid */
void app_remove (int aid) {
#ifdef DEBUG
  printf ("begin remove - proc %d\n", aid);
  dump_shmem ();
#endif

  mytp_lock (& (((daemon_t*)shmvaddr)->daemonlock));

  ((daemon_t*)shmvaddr)->sys_total -= ((daemon_t*)shmvaddr)->app[aid].total;
  ((daemon_t*)shmvaddr)->sys_running -= ((daemon_t*)shmvaddr)->app[aid].running;
  ((daemon_t*)shmvaddr)->app[aid].total = 0;	/* remove the registry entry */

  if ((((daemon_t*)shmvaddr)->sys_running < ((daemon_t*)shmvaddr)->sys_total) &&
      (((daemon_t*)shmvaddr)->sys_running < NUM_PROCS))
    redistribution();

  mytp_unlock (& (((daemon_t*)shmvaddr)->daemonlock));

#ifdef DEBUG
  printf ("end remove\n");
  dump_shmem ();
#endif

  shmem_end ();
}

/* probe to see if needs to suspend or resume suspended kernal thread */
int app_probe (int aid) {
  int ret;
  app_t *app_ptr = ((daemon_t*)shmvaddr)->app + aid;

#ifdef DEBUG
  printf ("begin probe - proc %d\n", aid);
  dump_shmem ();
#endif

  if (app_ptr->recommended == app_ptr->running) return 0;

  mytp_lock (& (((daemon_t*)shmvaddr)->daemonlock));

  if (app_ptr->recommended > app_ptr->running) {
    ret = app_ptr->recommended-app_ptr->running;
    app_ptr->running = app_ptr->recommended;
    ((daemon_t*)shmvaddr)->sys_running += ret;
  }
  else 
    if (app_ptr->recommended < app_ptr->running) {
      ret = -1;
      app_ptr->running --;
      ((daemon_t*)shmvaddr)->sys_running --;
    }

  mytp_unlock (& (((daemon_t*)shmvaddr)->daemonlock));

#ifdef DEBUG
  printf ("end probe\n");
  dump_shmem ();
#endif

  return ret;
}
