/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "mytp.h"

#define MYTP_STKSIZE (0x800000)		/* 4M (0x400000) stack size for sweep */
					/* 128K (0x20000) stack size for others */
#define MYTP_PRIVATE_SIZE (1024*sizeof(void*))

/* `alignment' must be a power of 2. */
#define MYTP_STKALIGN(sp, alignment) \
  ((void *)((((qt_word_t)(sp)) + (alignment) - 1) & ~((alignment)-1)))

static void *xmalloc (unsigned size) {
  void *sto;

  sto = malloc (size);
  if (!sto) {
    perror ("malloc");
    exit (1);
  }
  return (sto);
}

/* Queue access functions. */

/* A queue is a doubly-linked list of threads.  The queue head is a
   designated dummy list element.   */

static int queuelock = 0;

typedef struct {
  mytp_t t;
  mytp_t *tail;
} mytp_q_t;

static mytp_q_t mytp_global_runq;	/* head of the global runq */


static void mytp_qinit () {
  mytp_global_runq.t.next = mytp_global_runq.tail = &(mytp_global_runq.t);
}


static mytp_t *mytp_qget () {
  mytp_t *t;
  
  mytp_lock (&queuelock);	/* aquire the queue lock */

  t = mytp_global_runq.t.next;
  mytp_global_runq.t.next = t->next;
  if (t->next == &(mytp_global_runq.t)) {
    if (t == &(mytp_global_runq.t)) {	/* already empty */
      t = NULL;
    }
    else 
      mytp_global_runq.tail = &(mytp_global_runq.t);
  }
  
  mytp_unlock (&queuelock);	/* release the queue lock */

  return t;
}


static void mytp_qput (mytp_t *t) {
  mytp_lock (&queuelock);	/* aquire the queue lock */

  mytp_global_runq.tail->next = t;
  t->next = &(mytp_global_runq.t);
  mytp_global_runq.tail = t;

  mytp_unlock (&queuelock);	/* release the queue lock */
}


/* routines interacting with the daemon */
extern int app_register (int, int);
extern void app_remove (int);
extern int app_probe (int);


/* Thread routines. */

typedef struct {
  int myid;
} arg_t;

enum {RUNNING=0, SUSPEND=1};

static int nproc;	/* number of threads */
static int aid;		/* application ID */

static int vp_lock = 0;
static pid_t *pid;
static int *vp_status;

#define MAX_THR_NUM 64

static int thr_num=0;
static mytp_t *(thr_array[MAX_THR_NUM]);

enum {mytpReady=0, mytpRunning=1, mytpBlock=2, mytpDone=3};

static void *mytp_starthelp (qt_t *old, void *ignore0, void *ignore1);
static void mytp_only (void *pu, void *pt, qt_userf_t *f);
static void *mytp_aborthelp (qt_t *sp, void *old, void *null);
static void *mytp_yieldhelp (qt_t *sp, void *old, void *blockq);

static int start_abort=0;

void mytp_init () {
  mytp_qinit ();
}

void mytp_body (void *arg) {
  mytp_t *next;
  pid_t mypid = getpid();
  int i, change, resumed;
  int myid = ((arg_t*)arg)->myid;

  if (myid==0) pid[0]=mypid;
  vp_status[myid] = RUNNING;
  mytp_vp_local->vp_id = myid;

  for (;;) {
    change = app_probe (aid);
    if (change<0) {	/* suspend myself */
      mytp_lock (&vp_lock);
      vp_status[myid] = SUSPEND;
      mytp_unlock (&vp_lock);
      if (blockproc (mypid) != 0) {
	perror ("blockproc");
	exit (1);
      }
    }
    if (change>0) {
      mytp_lock (&vp_lock);
      i=0;
      resumed=0;
      while (resumed<change) {
	if (vp_status[i] == SUSPEND) {
	  vp_status[i] = RUNNING;
	  unblockproc (pid[i]);
	  resumed++;
	}
	i++;
      }
      mytp_unlock (&vp_lock);
    }

    while ((next = mytp_qget ()) == NULL) {
      if (start_abort) goto OUTOFHERE;
      uspin (IT_1us * 10);
    }

    mytp_vp_local->vp_curr = next;
    next->state = mytpRunning;
    QT_BLOCK (mytp_starthelp, NULL, NULL, next->sp);
  }

 OUTOFHERE:
  /* resume a suspended vp before exiting */
  mytp_lock (&vp_lock);

  for (i=0; i<nproc; i++)
    if (vp_status[i] == SUSPEND) {
      vp_status[i] = RUNNING;
      unblockproc (pid[i]);
      break;
    }
  mytp_unlock (&vp_lock);
}

static void *mytp_starthelp (qt_t *old, void *ignore0, void *ignore1) {
  mytp_vp_local->vp_main.sp = old;
  return NULL;
}

/* mytp assumes that all threads are created before calling mytp_start */
void mytp_start (int num_proc) {
  int i;
  arg_t* arg;

  nproc = num_proc;
  pid = (pid_t*) malloc (nproc*sizeof(pid_t));
  vp_status = (int*) malloc (nproc*sizeof(int));

  aid = app_register (nproc, 0);

  for (i=1; i<nproc; i++) {
    arg = (arg_t*) malloc(sizeof(arg_t));
    arg->myid = i;
    pid[i] = sproc (mytp_body, PR_SADDR, arg);
  }

  arg = (arg_t*) malloc(sizeof(arg_t));
  arg->myid = 0;
  mytp_body (arg);

  for (i=1; i<nproc; i++)
    waitpid (pid[i], NULL, NULL);

  app_remove (aid);
}


int mytp_create (mytp_userf_t *f, void *pu) {
  mytp_t *t;
  void *sto;

  t = xmalloc (sizeof(mytp_t));
  if (thr_num == MAX_THR_NUM) {
    perror ("too many threads");
    exit (1);
  }
  t->thrid = thr_num;
  thr_array[thr_num++] = t;
  t->state = mytpReady;
  t->sto = xmalloc (MYTP_STKSIZE);
  sto = MYTP_STKALIGN (t->sto, QT_STKALIGN);
  t->sp = QT_SP (sto, MYTP_STKSIZE - QT_STKALIGN);
  t->sp = QT_ARGS (t->sp, pu, t, (qt_userf_t *)f, mytp_only);
  t->private = xmalloc (MYTP_PRIVATE_SIZE);
  bzero (t->private, MYTP_PRIVATE_SIZE);
  mytp_qput (t);

  return t->thrid;
}

#ifdef STAT
#define sync	(mytp_vp_local->vp_curr->sync)
#define yield0	(mytp_vp_local->vp_curr->yield0)
#define yield1	(mytp_vp_local->vp_curr->yield1)
#define yield2	(mytp_vp_local->vp_curr->yield2)
#define yield3	(mytp_vp_local->vp_curr->yield3)
#define yield4up	(mytp_vp_local->vp_curr->yield4up)
#define lastrun	(mytp_vp_local->vp_curr->lastrun)
#define migrate (mytp_vp_local->vp_curr->migrate)
#endif

static void mytp_only (void *pu, void *pt, qt_userf_t *f) {
#ifdef STAT
  sync=0;
  yield0=0;
  yield1=0;
  yield2=0;
  yield3=0;
  yield4up=0;
  lastrun=-1;
  migrate=-1;
#endif
  ((mytp_t*)pt)->state = mytpRunning;
  mytp_vp_local->vp_curr = (mytp_t *)pt;
  (*(mytp_userf_t *)f)(pu);
  ((mytp_t*)pt)->state = mytpDone;
#ifdef STAT
  printf ("thr %d: sync-%d, yield0-%d, yield1-%d, yield2-%d, yield3-%d, yield4up-%d, migrate-%d\n", mytp_get_myid(), sync, yield0, yield1, yield2, yield3, yield4up, migrate);
  fflush (stdout);
#endif
  mytp_abort();
  /* NOTREACHED */
}


void mytp_abort (void) {
  mytp_t *old, *mainthread;

  start_abort = 1;

  mainthread = &(mytp_vp_local->vp_main);
  old = mytp_vp_local->vp_curr;
  old->state = mytpDone;
  mytp_vp_local->vp_curr = mainthread;
  QT_ABORT (mytp_aborthelp, old, (void *)NULL, mainthread->sp);
}


static void *mytp_aborthelp (qt_t *sp, void *old, void *null) {
  free (((mytp_t *)old)->sto);
  free (old);
  return NULL;
}


void mytp_yield () {
  mytp_t *old, *new;

  if (app_probe (aid) == 0) {
    if ((new = mytp_qget ())==NULL) return;
  }
  else {
    new = &(mytp_vp_local->vp_main);	/* go back to main thread */
  }

#ifdef STAT
  if (mytp_vp_local->vp_id != lastrun) {
    migrate++;
    lastrun = mytp_vp_local->vp_id;
  }
#endif

  old = mytp_vp_local->vp_curr;
  old->state = mytpBlock;
  mytp_vp_local->vp_curr = new;
  new->state = mytpRunning;
  QT_BLOCK (mytp_yieldhelp, old, NULL, new->sp);
}


static void *mytp_yieldhelp (qt_t *sp, void *old, void *blockq) {
  ((mytp_t *)old)->sp = sp;
  mytp_qput ((mytp_t *)old);
  return NULL;
}

#define ROUND 100

/* synchronization functions */
#define OP(a,b) (a==b)
int waitFlagEQ (int *pflag, int value, int peer) {
  int i;
#ifdef COLL2
  int lastrunon = mytp_vp_local->vp_id;
#endif
#ifdef STAT
  int yield_num=0;
  sync++;
#endif

  for (;;) {
    for (i=0; i<ROUND; i++) {
      if OP(*pflag,value) goto EXIT;
#ifdef COLL1
      if (peer>=0 && ((int)(thr_array[peer]->state))!=mytpRunning) goto YIELD;
#endif
      uspin (IT_1us);	 /* wait 1us */
    }
    if OP(*pflag,value) goto EXIT;
  YIELD:
#ifdef STAT
    yield_num++;
#endif
    mytp_yield ();	 /* yield */
#ifdef COLL2
    if (mytp_vp_local->vp_id!=lastrunon) {
      if OP(*pflag,value) goto EXIT;
      else goto YIELD;
    }
#endif
  }
 EXIT:
#ifdef STAT
  switch (yield_num) {
  case 0: yield0++; break;
  case 1: yield1++; break;
  case 2: yield2++; break;
  case 3: yield3++; break;
  default: yield4up++;
  }
#endif
  return 0;
}

#undef OP

#define OP(a,b) (a!=b)
int waitFlagNE (int *pflag, int value, int peer) {
  int i;
#ifdef COLL2
  int lastrunon = mytp_vp_local->vp_id;
#endif
#ifdef STAT
  int yield_num=0;
  sync++;
#endif

  for (;;) {
    for (i=0; i<ROUND; i++) {
      if OP(*pflag,value) goto EXIT;
#ifdef COLL1
      if (peer>=0 && ((int)(thr_array[peer]->state))!=mytpRunning) goto YIELD;
#endif
      uspin (IT_1us);	 /* wait 1us */
    }
    if OP(*pflag,value) goto EXIT;
  YIELD:
#ifdef STAT
    yield_num++;
#endif
    mytp_yield ();	 /* yield */
#ifdef COLL2
    if (mytp_vp_local->vp_id!=lastrunon) {
      if OP(*pflag,value) goto EXIT;
      else goto YIELD;
    }
#endif
  }
 EXIT:
#ifdef STAT
  switch (yield_num) {
  case 0: yield0++; break;
  case 1: yield1++; break;
  case 2: yield2++; break;
  case 3: yield3++; break;
  default: yield4up++;
  }
#endif
  return 0;
}

#undef OP

#define OP(a,b) (a&b)
int waitFlagIN (int *pflag, int value, int peer) {
  int i;
#ifdef COLL2
  int lastrunon = mytp_vp_local->vp_id;
#endif
#ifdef STAT
  int yield_num=0;
  sync++;
#endif

  for (;;) {
    for (i=0; i<ROUND; i++) {
      if OP(*pflag,value) goto EXIT;
#ifdef COLL1
      if (peer>=0 && ((int)(thr_array[peer]->state))!=mytpRunning) goto YIELD;
#endif
      uspin (IT_1us);	 /* wait 1us */
    }
    if OP(*pflag,value) goto EXIT;
  YIELD:
#ifdef STAT
    yield_num++;
#endif
    mytp_yield ();	 /* yield */
#ifdef COLL2
    if (mytp_vp_local->vp_id!=lastrunon) {
      if OP(*pflag,value) goto EXIT;
      else goto YIELD;
    }
#endif
  }
 EXIT:
#ifdef STAT
  switch (yield_num) {
  case 0: yield0++; break;
  case 1: yield1++; break;
  case 2: yield2++; break;
  case 3: yield3++; break;
  default: yield4up++;
  }
#endif
  return 0;
}


#undef OP

#define OP(a,b) (!(a%b))
int waitFlagMOD (int *pflag, int value, int peer) {
  int i;
#ifdef COLL2
  int lastrunon = mytp_vp_local->vp_id;
#endif
#ifdef STAT
  int yield_num=0;
  sync++;
#endif

  for (;;) {
    for (i=0; i<ROUND; i++) {
      if OP(*pflag,value) goto EXIT;
#ifdef COLL1
      if (peer>=0 && ((int)(thr_array[peer]->state))!=mytpRunning) goto YIELD;
#endif
      uspin (IT_1us);	 /* wait 1us */
    }
    if OP(*pflag,value) goto EXIT;
  YIELD:
#ifdef STAT
    yield_num++;
#endif
    mytp_yield ();	 /* yield */
#ifdef COLL2
    if (mytp_vp_local->vp_id!=lastrunon) {
      if OP(*pflag,value) goto EXIT;
      else goto YIELD;
    }
#endif
  }
 EXIT:
#ifdef STAT
  switch (yield_num) {
  case 0: yield0++; break;
  case 1: yield1++; break;
  case 2: yield2++; break;
  case 3: yield3++; break;
  default: yield4up++;
  }
#endif
  return 0;
}
