/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

/**
 * Parallel Gaussian Elimination - POSIX thread version
 * Hong Tang
 * 4/22/97
 */

#include <stdio.h>
#include "comm.h"
#include "util.h"

#define RESULT 1 /* to check whether the solution is correct, turn it to 1 */
/**
  * This program solves the matrix equation Ax=b by using block based GE with pivoting.
  * A: double[n][n]; x,b: double[n]. The matrix A is initialized in the following form:
  * all elements are 1 except those along the diagonal from top right to bottom left 
  * are n. All elements in b are 2n-1.  Thus x must contains all 1's.
  * 
  * The structure of the program is essentially the same as the MPI version except 
  * that all global variables are eliminated and all MPI primitives are replaced by
  * pthread based functions (defined in util.c).
  */


# define n MATRIX_SIZE /* dimension of matrix A */
int p; /* number of processors */
# define r CACHE_BLOCK /* dimension of the sub-matrix; 3r^2 <= cache size */
# define s (n/r) /* dimension of the matrix in the unit of sub-matrix */
int q; /* q=s/p: each processor owns s X q sub-matrices */
/* n must be dividable by pr. */

/* macro to convert two dimension array access to one dimension form */
# define E(M, i, j) (M)[i*r+j] 

int N=r, ione=1;
double negone = -1.0, one=1.0;

MBOX **mboxes;
BARRIER gbar;
BCOBJ **gbco1, **gbco2; /* gbco1 for piv, gbco2 for lub */

typedef struct {
	int me; /* id of self */
	double *(*BLK)[s];
	double *b;
	double *tmp[s];
	int piv[r]; /* pivoting information */
} GV; /* structure for all global variables */

typedef struct {
	int id;
} PARM;


# define me (gv->me)
# define BLK (gv->BLK)
# define b (gv->b)
# define tmp (gv->tmp)
# define piv (gv->piv)

void LU(GV *gv)
{
	int BCol, BRow, Col, Row;
	double max;
	double *block, *lub;
	double *blk2;
	int imax;
	int i,j;
	double dtmp;
	int start;

	for (BCol=0; BCol<s; BCol++)  {
		if ((BCol%p)==me) {  /* do LU on BLK[BCol/p][BCol] */
			lub=BLK[BCol/p][BCol];
			for (Col=0;Col<r;Col++) {
				max=E(lub, Col,Col);
				imax=BCol*r+Col;
				block=lub;
				for (Row=Col+1;Row<r;Row++)
					if (max<E(block, Col, Row)) {
						max=E(block, Col, Row);
						imax=BCol*r+Row;
					}
				for (BRow=BCol+1; BRow<s;BRow++) {
					block=BLK[BCol/p][BRow];
					for (Row=0; Row<r; Row++)
						if (max<E(block, Col, Row)) {
							max=E(block, Col, Row);
							imax=BRow*r+Row;
						}
				}
				if (imax!=BCol*r+Col) {
					piv[Col]=imax;
					/* swapping */
					block=BLK[BCol/p][imax/r];
					for (i=Col; i<r;i++) {
						dtmp=E(block,i,imax%r);
						E(block,i,imax%r)=E(lub,i,Col);
						E(lub,i,Col)=dtmp;
					}
					for (j=BCol/p+1;j<q;j++) {
						block=BLK[j][imax/r];
						blk2=BLK[j][BCol];
						for (i=0;i<r;i++) {
							dtmp=E(block,i,imax%r);
							E(block,i,imax%r)=E(blk2,i,Col);
							E(blk2,i,Col)=dtmp;
						}
					}
					if (me==p-1) { /* swap b */
						dtmp=b[BCol*r+Col];
						b[BCol*r+Col]=b[imax];
						b[imax]=dtmp;
					}
				}
				else 
					piv[Col]=-1;

				/* scaling and updating */
				block=lub;
				for (Row=Col+1;Row<r;Row++) {
					E(block,Col,Row)/=max;
					for (i=Col+1;i<r;i++)
						E(block,i,Row)-=E(lub,i,Col)*E(block,Col,Row);
					/**
					if (me==p-1)
						b[BCol*r+Row]-=b[BCol*r+Col]*E(block,Col,Row);
					**/
				}
				for (BRow=BCol+1; BRow<s;BRow++) {
					block=BLK[BCol/p][BRow];
					for (Row=0; Row<r; Row++) {
						E(block,Col,Row)/=max;
						for (i=Col+1;i<r;i++)
							E(block,i,Row)-=E(lub,i,Col)*E(block,Col,Row);
						/***
						if (me==p-1)
							b[BRow*r+Row]-=b[BCol*r+Col]*E(block,Col,Row);
						***/
					}
				}
			}
			/* sending out pivoting info */
			if (BCol<s-1) {
				if (p>1) {
					bco_aggbcast(gbco1[me], (char *)piv, r*sizeof(int), me, me );
					/* sending out L and column blocks info */
					bco_aggbcast(gbco2[me], (char *)lub, r*r*(s-BCol)*sizeof(double), me, me);
					/* updating blocks below and to the right of lub */
				}
				for (i=BCol/p+1;i<q;i++) {
					block=BLK[i][BCol];
					dtrsm_("L","L","N","U",&N,&N,&one,lub, &N, block, &N);
					for (BRow=BCol+1; BRow<s;BRow++) {
						blk2=BLK[i][BRow];
						dgemm_("N","N",&N,&N,&N,&negone,BLK[BCol/p][BRow],&N,block,&N,&one,blk2,&N);
					}
				}
			}

			if (me==p-1) { /* update b */
				dtrsv_("L","N","U",&N,lub,&N,&(b[BCol*r]),&ione);
				for (BRow=BCol+1;BRow<s;BRow++) 
					dgemv_("N",&N,&N,&negone,BLK[BCol/p][BRow],&N,&(b[BCol*r]),&ione,&one,\
						&(b[BRow*r]),&ione);
			}
		}
		else {
			if (BCol<s-1) {
				if (p>1) {
					/* receive piv */
					bco_aggbcast(gbco1[BCol%p], (char *)piv, r*sizeof(int), BCol%p, me );
					/* receive column blocks */
					bco_aggbcast(gbco2[BCol%p], (char *)(tmp[BCol]), r*r*(s-BCol)*sizeof(double), BCol%p, me);
				}
				/* do pivoting first */

				if (me<(BCol%p))
					start=1;
				else
					start=0;

				for (Row=0;Row<r;Row++) {
					if (piv[Row]!=-1) { /* need swapping */
						for (i=BCol/p+start;i<q;i++) {
							blk2=BLK[i][BCol];
							block=BLK[i][piv[Row]/r];
							for (Col=0;Col<r;Col++) {
								dtmp=E(blk2,Col,Row);
								E(blk2,Col,Row)=E(block,Col,piv[Row]%r);
								E(block,Col,piv[Row]%r)=dtmp;
							}
						}
						if (me==p-1) { /* swap b */
							dtmp=b[BCol*r+Row];
							b[BCol*r+Row]=b[piv[Row]];
							b[piv[Row]]=dtmp;
						}
					}
				}

				/* Aij=L^{-1}Aij and Ajk-=AjiAik */
				for (i=BCol/p+start;i<q; i++) {
					block=BLK[i][BCol];
					dtrsm_("L","L","N","U",&N,&N,&one,tmp[BCol],&N, block, &N);
					for (BRow=BCol+1; BRow<s;BRow++) {
						blk2=BLK[i][BRow];
						dgemm_("N","N",&N,&N,&N,&negone,tmp[BRow],&N,block,&N,&one,blk2,&N);
					}
				}
				if (me==p-1) {
					dtrsv_("L","N","U",&N,tmp[BCol],&N,&(b[BCol*r]),&ione);
					for (BRow=BCol+1;BRow<s;BRow++) 
						dgemv_("N",&N,&N,&negone,tmp[BRow],&N,&b[BCol*r],&ione,&one,&b[BRow*r],&ione);
				}
			}
		}
	}
}

void Init(GV *gv)
{
	int BCol, BRow, Row, Col;

	for (BCol=0;BCol<q;BCol++)
		for (BRow=0;BRow<s;BRow++) /* initializing BLK[Bcol][BRow] */
		{
			for (Col=0;Col<r;Col++)
				for(Row=0;Row<r;Row++)
					E(BLK[BCol][BRow],Col,Row)=1;

			if (BCol*p+me+BRow+1==s) {
				for (Col=0;Col<r;Col++)
					E(BLK[BCol][BRow],Col,r-Col-1)=n;
			}
		}

	/* the last processor has the responsibility of holding and updating b */
	if( p-1 == me )
	{
		for( Row=0; Row<n; Row++ )
			b[Row] = 2*n-1;
	}
}

void Back(GV *gv)
{
	int BlockedColumn, BlockedRow;
	int BCol, BRow;

	for (BCol=s-1;BCol>=0;BCol--) {
		if (BCol%p==me) {
			dtrsv_("U","N","N",&N,BLK[BCol/p][BCol],&N,&b[BCol*r],&ione);
			for(BRow=BCol-1; BRow>=0; BRow--)
				dgemv_("N",&N,&N,&negone,BLK[BCol/p][BRow],&N,&b[BCol*r],&ione,&one,&b[BRow*r],&ione);

			if( (BCol>0) && (p>1)) 
				mbox_asend(mboxes[(me-1+p)%p], (char *)b, n*sizeof(double), me);
		}
		else if ((BCol>0) && ((BCol-1+p)%p == me) && (p>1) )
			mbox_arecv(mboxes[me], (char *)b, n*sizeof(double), (me+1)%p);
	}
}

void result(GV *gv)
{
	int Row;

	if( me==0 ) {
		for( Row=0; Row<n; Row++ )
			if (b[Row]<0.999999 || b[Row]>1.000001)
				printf( "x[%d] = %e\n", Row, b[Row] );
	}
}

void display_matrix(GV *gv)
{
	int BCol, BRow, Row, Col;
	int i;
	double *blk;

	for (i=0;i<q;i++) {
		BCol=i*p+me;
		for (BRow=0;BRow<s;BRow++) {
			blk=BLK[i][BRow];
			for (Col=0;Col<r;Col++)
				for (Row=0;Row<r;Row++)
					printf("A[%d][%d]=%e\n",BRow*r+Row, BCol*r+Col, E(blk,Col, Row));
		}
	}
}

# undef me
# undef BLK
# undef b
# undef tmp
# undef piv

#ifdef PTHREAD
void * tmain
#endif
#ifdef MYTP
void tmain
#endif
( void * arg)
{
	PARM *parm=(PARM *)arg;

	double t1, t2;
	int i;

	GV gv;

	mboxes[parm->id]=(MBOX *)malloc(MAX(sizeof(MBOX), 128));
	mbox_init(mboxes[parm->id]);

	gbco1[parm->id]=(BCOBJ *)malloc(MAX(sizeof(BCOBJ), 128));
	bco_init(gbco1[parm->id], p);

	gbco2[parm->id]=(BCOBJ *)malloc(MAX(sizeof(BCOBJ), 128));
	bco_init(gbco2[parm->id], p);

	bar_wait(&gbar);

	gv.me=parm->id;
	gv.BLK=(double *(*)[s])malloc(sizeof(double*[s])*q);
	gv.BLK[0][0]=(double *)malloc(sizeof(double)*n*n/p);
	for (i=1;i<q*s;i++)
		gv.BLK[0][i]=gv.BLK[0][i-1]+r*r;

	gv.b=(double *)malloc(sizeof(double)*n);
	gv.tmp[0]=(double *)malloc(sizeof(double)*r*n);
	for (i=1;i<s;i++)
		gv.tmp[i]=gv.tmp[i-1]+r*r;

	Init(&gv);

	/* display_matrix(&gv); */

	/* start the timer */
	if( gv.me==0 ) {
		t1 = get_wtime();
		printf("Solving Ax=b ... \n");
	}

	LU(&gv);
	/*
	if (gv.me==0)
		t2 = get_wtime();
	*/

	/* printf("(%d) before Back()\n",gv.me); */

	/* back substitution */
	Back(&gv);

	if( gv.me==0 ) { /* 0 will be the last to finish back substitution */

		/* stop the timer */
		t2 = get_wtime();

		/* display the elapsed time */
		printf( "CPU time spent is %e seconds.\n", t2-t1 );
	}

	/* everyone must wait for the result and time to be displayed */
	bar_wait(&gbar);
# if (RESULT)
	result(&gv);
# endif
#ifdef PTHREAD	
	return NULL;
#endif
}

int main( int argc, char *argv[] )
{
	int i;
	int num_vp;
	PARM *parm;
#ifdef PTHREAD 
	pthread_t *threads;
#endif
	if (argc < 3)
	{
		printf("Usage: %s m n\n  where m is the degree of concurrency and n is no. of thread\n", argv[0]);
		exit(1);
	}

	num_vp = atoi(argv[1]);
	p = atoi(argv[2]);

	if (n%(p*r)) {
		printf("n must be divivable by (25p), where p"
		    " is the number of processors\n");
		exit(0);
	}
#ifdef PTHREAD
	pthread_setconcurrency(num_vp);
#endif
	q=s/p;

	parm=(PARM *)malloc(p*sizeof(PARM));
#ifdef PTHREAD
	threads=(pthread_t *)malloc(p*sizeof(pthread_t));
#endif
	mboxes=(MBOX **)malloc(p*sizeof(MBOX *));
	gbco1=(BCOBJ **)malloc(p*sizeof(BCOBJ*));
	gbco2=(BCOBJ **)malloc(p*sizeof(BCOBJ*));

	bar_init(&gbar, p);
	/*
	for (i=0;i<p;i++) {
		bco_init(&gbco1[i], p);
		bco_init(&gbco2[i], p);
	}
	*/

  /* Spawn thread */
#ifdef MYTP
  mytp_init ();
#endif

	for (i=0;i<p;i++) {
		parm[i].id=i;
#ifdef PTHREAD
		pthread_create(&threads[i],NULL, tmain, (void *)(parm+i));
#endif
#ifdef MYTP
		mytp_create (tmain, (void *)(parm+i));
#endif
	}

#ifdef MYTP
	mytp_start (num_vp);
#endif
#ifdef PTHREAD
	for (i=0;i<p;i++) {
		pthread_join(threads[i], NULL);
	}
#endif
	return 0;
}
