/**
 * Copyright (c) 2000
 * University of California, Santa Barbara
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice and this notice
 * appear in all copies.  This software is provided as a
 * proof-of-concept and for demonstration purposes; there is no
 * representation about the suitability of this software for any
 * purpose.
 */

# include <errno.h>
#include <sys/time.h>

# include "util.h"

/***BARRIER OBJECT INTERFACE ***/

/* bar_init -  init the barrier */
int bar_init(BARRIER *bp, int count)
{
	int     n;
	int     i;

	if (count<1)
		return(EINVAL);

	bp->maxcnt = count;
	bp->sbp = &bp->sb[0];

	for (i=0; i<2; ++i) {
		struct _sb *sbp=&(bp->sb[i]);
		sbp->runners = count;

		if (n=pthread_mutex_init(&sbp->wait_lk, NULL))
			return(n);

		if (n=pthread_cond_init(&sbp->wait_cv, NULL))
			return(n);
	}
	return (0);
}

/*
 * bar_wait - wait at a barrier for everyone to arrive.
 *
 */

int bar_wait(BARRIER *bp)
{
	struct _sb *sbp = bp->sbp;

	pthread_mutex_lock( &sbp->wait_lk );

	if ( sbp->runners == 1 ) {      /* last thread to reach barrier */
		if ( bp->maxcnt != 1 ) {
			/* reset runner count and switch sub-barriers   */
			sbp->runners=bp->maxcnt;
			bp->sbp=(bp->sbp == &bp->sb[0] )? &bp->sb[1] : &bp->sb[0];

			/* wake up the waiters */
			pthread_cond_broadcast( &sbp->wait_cv );
		}
	} else {
		sbp->runners--;         /* one less runner */

		while ( sbp->runners != bp->maxcnt ) /* important! */
			pthread_cond_wait( &sbp->wait_cv, &sbp->wait_lk );
	}

	pthread_mutex_unlock( &sbp->wait_lk );

	return (0);
}

/* bar_destroy - destroy a barrier variable. */

int bar_destroy( BARRIER *bp )
{
	int n;
	int i;

	for (i=0; i<2; ++i ) {
		if ( n = pthread_cond_destroy( &bp->sb[i].wait_cv ) )
			return (n);

		if ( n = pthread_mutex_destroy( &bp->sb[i].wait_lk ) )
			return (n);
	}

	return( 0 );
}

/*** simple mailbox object interface */
/* mbox_init - init the MBOX */
int mbox_init(MBOX *mbox)
{
	int n;

	if (n=pthread_mutex_init(&mbox->mutex, NULL))
		return(n);

	if (n=pthread_mutex_init(&mbox->empty, NULL))
		return(n);

	if (n=pthread_mutex_init(&mbox->full, NULL))
		return(n);

	pthread_mutex_lock(&mbox->full);

	if (n=pthread_mutex_init(&mbox->fnsh, NULL))
		return(n);

	pthread_mutex_lock(&mbox->fnsh);

	return 0;
}

/* mbox_destroy - destroy the MBOX */
int mbox_destroy(MBOX *mbox)
{
	int n;

	if (n=pthread_mutex_destroy(&mbox->mutex))
		return(n);

	if (n=pthread_mutex_destroy(&mbox->empty))
		return(n);

	if (n=pthread_mutex_destroy(&mbox->full))
		return(n);

	if (n=pthread_mutex_destroy(&mbox->fnsh))
		return(n);
	
	return 0;
}

/* mbox_recv - synchronized receive */
int mbox_recv(MBOX *mbox, char *data, int size, int from)
{
	int n;
	char *src;

	pthread_mutex_lock(&mbox->full);

	if (mbox->from==from) {
		n=0;
		src=mbox->data;
	}
	else {
		n=-1;
	}

	pthread_mutex_unlock(&mbox->empty);
	pthread_mutex_unlock(&mbox->fnsh);

	memcpy(data, src, MIN(size, mbox->size));

	return n;
}

/* mbox_arecv - still synchronized receive, but is used to match mbox_asend */
int mbox_arecv(MBOX *mbox, char *data, int size, int from)
{
	int n;
	char *src;

	pthread_mutex_lock(&mbox->full);

	if (mbox->from==from) {
		n=0;
		src=mbox->data;
		/* memcpy(data, mbox->data, MIN(size, mbox->size)); */
	}
	else {
		n=-1;
	}

	pthread_mutex_unlock(&mbox->empty);

	memcpy(data, src, MIN(size, mbox->size));

	return n;
}

/* mbox_asend - asynchronized send, from is the id of the sender */
int mbox_asend(MBOX *mbox, char *data, int size, int from)
{
	pthread_mutex_lock(&mbox->empty);

	mbox->from=from;
	mbox->data=data;
	mbox->size=size;

	pthread_mutex_unlock(&mbox->full);

	return 0;
}

/*** broadcast object interface */
/* bco_init - init the BCOBJ */
int bco_init(BCOBJ *bco, int maxcnt)
{
	int n;

	bco->avail=0;

	if (n=pthread_mutex_init(&bco->mutex, NULL))
		return(n);

	if (n=pthread_cond_init(&bco->cond, NULL))
		return(n);

	if (n=bar_init(&bco->bar, maxcnt))
		return(n);

	return 0;
}

/* bco_destroy - destroy the BCOBJ */
int bco_destroy(BCOBJ *bco)
{
	int n;

	if (n=pthread_mutex_destroy(&bco->mutex))
		return(n);

	if (n=pthread_cond_destroy(&bco->cond))
		return(n);

	if (n=bar_destroy(&bco->bar))
		return(n);

	return 0;
}

/* bco_bcast - broadcast the data from the sender to all other threads */
int bco_bcast(BCOBJ *bco, char *data, int size, int snder, int me)
{
	char *src;

	if (snder!=me) {
		pthread_mutex_lock(&bco->mutex);
		while (!bco->avail)
			pthread_cond_wait(&bco->cond, &bco->mutex);
		
		/* this ensures avail goes to 0 after all the receivers get the data */
		bco->avail=(bco->avail+1)% bco->bar.maxcnt;
		src=bco->data;
	}
	else {
		pthread_mutex_lock(&bco->mutex);
		bco->data=data;
		bco->avail=1;
		pthread_cond_broadcast(&bco->cond);
	}
	pthread_mutex_unlock(&bco->mutex);

	bar_wait(&bco->bar);
	if (snder!=me)
		memcpy(data, src, size);
	return 0;
}

/* bco_aggbcast - aggressive broadcast the data from the sender to all other threads */
/* to make it safe, there must be some synchronization point between two bco_aggbcast() */
/* calls on the same bco. To simplify a problem, use different bco's for bco_aggbcast() */
/* calls that are too near to each other. */
int bco_aggbcast(BCOBJ *bco, char *data, int size, int snder, int me)
{
	char *src;

	pthread_mutex_lock(&bco->mutex);
	if (snder!=me) {
		while (!bco->avail)
			pthread_cond_wait(&bco->cond, &bco->mutex);
		
		/* this ensures avail goes to 0 after all the receivers get the data */
		bco->avail=(bco->avail+1)% bco->bar.maxcnt;
		src=bco->data;
		/* memcpy(data, bco->data, size); */
	}
	else {
		bco->data=data;
		bco->avail=1;
		pthread_cond_broadcast(&bco->cond);
	}
	pthread_mutex_unlock(&bco->mutex);

	if (snder!=me)
		memcpy(data, src, size);
	return 0;
}

double get_wtime()
{
	struct timeval t;

	gettimeofday(&t, NULL);

	return (double)t.tv_sec+t.tv_usec*0.000001;
}
